# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Batched environment."""

from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import numpy as np
from gym import Env
from numpy import ndarray

from actorch.envs.utils import Nested, batch_space, unbatch


__all__ = [
    "BatchedEnv",
]


class BatchedEnv(ABC, Env):
    """Environment that runs multiple copies of a base environment."""

    def __init__(
        self, env_builder: "Callable[[], Env]", num_workers: "int" = 1
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        env_builder:
            The base environment builder.
        num_workers:
            The number of copies of the base environment.

        Raises
        ------
        ValueError
            If `num_workers` is not in the integer interval [1, inf).

        """
        if num_workers < 1 or not float(num_workers).is_integer():
            raise ValueError(
                f"`num_workers` ({num_workers}) must be in the integer interval [1, inf)"
            )
        num_workers = int(num_workers)
        self.env_builder = env_builder
        self.num_workers = num_workers
        dummy_env = env_builder()
        self.single_observation_space = dummy_env.observation_space
        self.single_action_space = dummy_env.action_space
        self.metadata = dummy_env.metadata
        self.reward_range = dummy_env.reward_range
        self.spec = dummy_env.spec
        self._repr = (
            f"<{self.__class__.__name__}"
            f"(env: {dummy_env}, "
            f"num_workers: {num_workers})>"
        )
        dummy_env.close()
        del dummy_env
        self.observation_space = batch_space(self.single_observation_space, num_workers)
        self.action_space = batch_space(self.single_action_space, num_workers)
        self._is_closed = False

    # override
    def reset(
        self,
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
    ) -> "Nested[ndarray]":
        """Reset a batch of workers.

        In the following, let `N` denote the number of workers
        and `O` the shape of a single observation leaf value.

        Parameters
        ----------
        mask:
            The boolean array indicating which workers are
            to reset (True) and which are not (False), shape: ``[N]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(N, dtype=bool)``.

        Returns
        -------
            The batched initial observation,
            shape of a leaf value: ``[N, *O]``.

        Raises
        ------
        RuntimeError
            If the environment has already been closed.
        ValueError
            If length of `mask` is not equal to the number of workers.

        Warnings
        --------
        If ``mask[i]`` is ``False``, the corresponding
        observation preserves its previous value.

        """
        if self._is_closed:
            raise RuntimeError("Trying to reset a closed environment")
        if mask is None:
            mask = np.ones(self.num_workers, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (self.num_workers,))
            if len(mask) != self.num_workers:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be equal "
                    f"to the number of workers ({self.num_workers})"
                )
        idx = np.where(mask)[0].tolist()
        return self._reset(idx)

    # override
    def step(
        self,
        action: "Nested[ndarray]",
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
    ) -> "Tuple[Nested[ndarray], ndarray, ndarray, ndarray]":
        """Step a batch of workers.

        In the following, let `N` denote the number of workers,
        `O` the shape of a single observation leaf value and
        `A` the shape of a single action leaf value.

        Parameters
        ----------
        action:
            The batched action, shape of a leaf value: ``[N, *A]``.
        mask:
            The boolean array indicating which workers are
            to step (True) and which are not (False), shape: ``[N]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(N, dtype=bool)``.

        Returns
        -------
            - The batched next observation, shape of a leaf value: ``[N, *O]``;
            - the batched reward, shape: ``[N]``;
            - the batched end-of-episode flag, shape: ``[N]``;
            - the batched auxiliary diagnostic information, shape: ``[N]``.

        Raises
        ------
        RuntimeError
            If the environment has already been closed.
        ValueError
            If length of `mask` or `action` is not equal
            to the number of workers.

        Warnings
        --------
        If ``mask[i]`` is ``False``, the corresponding
        observation, reward, end-of-episode flag and
        auxiliary diagnostic information preserve
        their previous value.

        """
        if self._is_closed:
            raise RuntimeError("Trying to step a closed environment")
        if mask is None:
            mask = np.ones(self.num_workers, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (self.num_workers,))
            if len(mask) != self.num_workers:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be equal "
                    f"to the number of workers ({self.num_workers})"
                )
        action = unbatch(self.action_space, action)
        if len(action) != self.num_workers:
            raise ValueError(
                f"Length of `action` ({len(action)}) must be equal "
                f"to the number of workers ({self.num_workers})"
            )
        idx = np.where(mask)[0].tolist()
        return self._step(action, idx)

    # override
    def seed(
        self,
        seed: "Union[Optional[int], Sequence[Optional[int]], ndarray]" = None,
    ) -> "None":
        """Seed all workers.

        In the following, let `N` denote the number of workers.

        Parameters
        ----------
        seed:
            The batched seed, shape ``[N]``.
            If a scalar or a scalar singleton, it
            is expanded to ``[seed + i for i in range(N)]`` or
            ``[seed[0] + i for i in range(N)]``, respectively.
            Default to ``[None] * N``.

        Raises
        ------
        RuntimeError
            If the environment has already been closed.
        ValueError
            If length of `seed` is not equal to the number of workers.

        """
        if self._is_closed:
            raise RuntimeError("Trying to seed a closed environment")
        if seed is None:
            seed = [None] * self.num_workers
        elif np.isscalar(seed):
            seed = [seed + i for i in range(self.num_workers)]
        elif len(seed) == 1:
            seed = (
                [None] * self.num_workers
                if seed[0] is None
                else [seed[0] + i for i in range(self.num_workers)]
            )
        elif len(seed) != self.num_workers:
            raise ValueError(
                f"Length of `seed` ({len(seed)}) must be equal "
                f"to the number of workers ({self.num_workers})"
            )
        elif isinstance(seed, np.ndarray):
            seed = seed.tolist()
        self.single_observation_space.seed(seed[0])
        self.observation_space.seed(seed[0])
        self.single_action_space.seed(seed[0])
        self.action_space.seed(seed[0])
        self._seed(seed)

    # override
    def render(
        self,
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
        **kwargs: "Any",
    ) -> "None":
        """Render a batch of workers.

        In the following, let `N` denote the number of workers.

        Parameters
        ----------
        mask:
            The boolean array indicating which workers are
            to render (True) and which are not (False), shape: ``[N]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(N, dtype=bool)``.

        Raises
        ------
        RuntimeError
            If the environment has already been closed.
        ValueError
            If length of `mask` is not equal to the number of workers.

        """
        if self._is_closed:
            raise RuntimeError("Trying to render a closed environment")
        if mask is None:
            mask = np.ones(self.num_workers, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (self.num_workers,))
            if len(mask) != self.num_workers:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be equal "
                    f"to the number of workers ({self.num_workers})"
                )
        idx = np.where(mask)[0].tolist()
        self._render(idx, **kwargs)

    # override
    def close(self) -> "None":
        """Close all workers."""
        if self._is_closed:
            return
        self._close()
        self._is_closed = True

    def __len__(self) -> "int":
        return self.num_workers

    def __del__(self) -> "None":
        if not self._is_closed:
            self.close()

    def __repr__(self) -> "str":
        return self._repr

    __str__ = __repr__

    @abstractmethod
    def _reset(self, idx: "Sequence[int]") -> "Nested[ndarray]":
        """See documentation of method `reset`."""
        raise NotImplementedError

    @abstractmethod
    def _step(
        self, action: "Sequence[Nested]", idx: "Sequence[int]"
    ) -> "Tuple[Nested[ndarray], ndarray, ndarray, ndarray]":
        """See documentation of method `step`."""
        raise NotImplementedError

    @abstractmethod
    def _seed(self, seed: "Sequence[Optional[int]]") -> "None":
        """See documentation of method `seed`."""
        raise NotImplementedError

    @abstractmethod
    def _render(self, idx: "Sequence[int]", **kwargs: "Any") -> "None":
        """See documentation of method `render`."""
        raise NotImplementedError

    @abstractmethod
    def _close(self) -> "None":
        """See documentation of method `close`."""
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Spaces."""

from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, Union

import numpy as np
from gym import spaces
from numpy import ndarray

from actorch.envs.utils import (
    Nested,
    batch_space,
    flatten,
    get_log_prob,
    get_space_bounds,
    unflatten,
    unnest_space,
)


__all__ = [
    "Flat",
]


class Flat(spaces.Box):
    """Space wrapper that flattens a base space."""

    def __init__(
        self,
        space: "spaces.Space",
        is_batched: "bool" = False,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        space:
            The base space to flatten.
        is_batched:
            True if `space` is batched, False otherwise.

        """
        # Check that space is batched when is_batched=True
        flatten(space, space.sample(), is_batched, copy=False, validate_args=True)
        self.space = space
        self.is_batched = is_batched
        low, high = get_space_bounds(space)
        low, high = self.flatten(low), self.flatten(high)
        unnested_space = unnest_space(space)
        self._shape_dict, self._type_dict = {}, {}
        for i, (key, space) in enumerate(unnested_space.items()):
            x = np.array(space.sample(), copy=False, ndmin=1)
            self._shape_dict[key] = x.shape
            self._type_dict[key] = type(space)
        super().__init__(low, high, dtype=low.dtype)
        self._cached_flat_unflat = None, None

    @property
    def shape_dict(self) -> "Dict[str, Tuple[int, ...]]":
        return self._shape_dict

    @property
    def type_dict(self) -> "Dict[str, Type[spaces.Space]]":
        return self._type_dict

    # override
    def sample(self) -> "ndarray":
        unflat = self.space.sample()
        flat = self.flatten(unflat)
        self._cached_flat_unflat = flat, unflat
        return flat

    # override
    def seed(self, seed: "Optional[int]" = None) -> "List[int]":
        return self.space.seed(seed)

    # override
    def contains(self, x: "ndarray") -> "bool":
        return self.space.contains(self.unflatten(x))

    # override
    def to_jsonable(
        self,
        sample_n: "Union[Sequence[ndarray], ndarray]",
    ) -> "List":
        return self.space.to_jsonable([self.unflatten(x) for x in sample_n])

    # override
    def from_jsonable(self, sample_n: "List") -> "ndarray":
        return np.asarray([self.flatten(x) for x in self.space.from_jsonable(sample_n)])

    def flatten(self, x: "Nested") -> "ndarray":
        """Flatten a sample from the underlying
        space (without copying if possible).

        Parameters
        ----------
        x:
            The sample.

        Returns
        -------
            The flat sample.

        """
        return flatten(self.space, x, self.is_batched, copy=False)

    def unflatten(self, x: "ndarray") -> "Nested":
        """Unflatten a flat sample from the underlying
        space (without copying if possible).

        Parameters
        ----------
        x:
            The flat sample.

        Returns
        -------
            The sample.

        """
        return unflatten(self.space, x, copy=False)

    def log_prob(self, x: "ndarray") -> "ndarray":
        """Return the log probability of a flat
        sample from the underlying space.

        Parameters
        ----------
        x:
            The flat sample.

        Returns
        -------
            The log probability.

        """
        flat, unflat = self._cached_flat_unflat
        if x is not flat:
            unflat = self.unflatten(x)
        return get_log_prob(self.space, unflat, self.is_batched)

    def __eq__(self, other: "Any") -> "bool":
        return (
            isinstance(other, Flat)
            and self.space == other.space
            and self.is_batched == other.is_batched
        )

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(space: {self.space}, "
            f"is_batched: {self.is_batched})"
        )


#####################################################################################################
# Flat batch_space implementation
#####################################################################################################


@batch_space.register(Flat)
def _batch_space_flat(space: "Flat", batch_size: "int" = 1) -> "Flat":
    return Flat(batch_space(space.space, batch_size), is_batched=True)


#####################################################################################################
# Flat get_log_prob implementation
#####################################################################################################


@get_log_prob.register(Flat)
def _get_log_prob_flat(
    space: "Flat",
    x: "ndarray",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if is_batched and not space.is_batched:
            raise ValueError(f"`x` ({x}) and `space` ({space}) must be batched")
    result = space.log_prob(x)
    return result.sum(axis=tuple(range(is_batched, result.ndim)))



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Reduction registry similar to PyTorch KL registry
(see https://github.com/pytorch/pytorch/blob/a347c747df8302acc0007a26f23ecf3355a5bef9/torch/distributions/kl.py#L36).

"""

from collections import defaultdict
from typing import Any, Callable, List, Optional, Tuple, Type, Union

import torch
from torch import distributions as ds

from actorch.distributions.cat_distribution import CatDistribution
from actorch.distributions.deterministic import Deterministic
from actorch.distributions.finite import Finite
from actorch.distributions.masked_distribution import MaskedDistribution
from actorch.distributions.transformed_distribution import TransformedDistribution
from actorch.distributions.transforms import SumTransform
from actorch.distributions.utils import is_affine, is_scaling
from actorch.utils import is_identical


__all__ = [
    "reduce",
    "register_reduction",
]


_REDUCTION_REGISTRY = defaultdict(list)

_REDUCTION_MEMOIZE = {}


def register_reduction(
    distribution_cls: "Type[ds.Distribution]",
    transform_cls: "Type[ds.Transform]",
    condition_fn: "Optional[Callable[[ds.Distribution, ds.Transform], bool]]" = None,
) -> "Callable":
    """Decorator to register a reduction function that maps a distribution-transform
    pair to its corresponding reduced distribution. Optionally, a condition function
    can be provided, that checks whether the reduction function is applicable to the
    given distribution-transform pair (useful, for example, to avoid ambiguities).

    Parameters
    ----------
    distribution_cls:
        The distribution class to register.
    transform_cls:
        The transform class to register.
    condition_fn:
        The function invoked before applying the reduction function.
        This function receives as arguments the distribution and the transform
        and returns True if the reduction function is applicable, False otherwise.

    Returns
    -------
        The decorated reduction function.

    Raises
    ------
    ValueError
        If an invalid argument value is provided.

    """
    if (not isinstance(distribution_cls, type)) or (
        not issubclass(distribution_cls, ds.Distribution)
    ):
        raise ValueError(
            f"`distribution_cls` ({distribution_cls}) must be a subclass of `torch.distributions.Distribution`"
        )
    if (not isinstance(transform_cls, type)) or (
        not issubclass(transform_cls, ds.Transform)
    ):
        raise ValueError(
            f"`transform_cls` ({transform_cls}) must be a subclass of `torch.distributions.Transform`"
        )

    def register_fn(
        reduction_fn: "Callable[..., ds.Distribution]",
    ) -> "Callable[..., ds.Distribution]":
        _REDUCTION_REGISTRY[distribution_cls, transform_cls].append(
            (condition_fn, reduction_fn)
        )
        _REDUCTION_MEMOIZE.clear()  # Reset since lookup order may have changed
        return reduction_fn

    return register_fn


def reduce(
    distribution: "ds.Distribution", transform: "ds.Transform", **kwargs: "Any"
) -> "ds.Distribution":
    """Reduce a distribution based on the applied transform.

    For example, if the distribution is a `Normal(mu, sigma)` and the applied transform
    is an `AffineTransform(loc, scale)`, then the resulting reduced distribution is a
    `Normal(loc + scale * mu, scale * sigma)`.

    Returns
    -------
        The reduced distribution.

    Raises
    ------
    NotImplementedError
        If no applicable reduction function exists.

    """
    try:
        reductions = _REDUCTION_MEMOIZE[type(distribution), type(transform)]
    except KeyError:
        reductions = _dispatch_reduction(type(distribution), type(transform))
        _REDUCTION_MEMOIZE[type(distribution), type(transform)] = reductions
    for condition_fn, reduction_fn in reductions:
        if (not condition_fn) or condition_fn(distribution, transform):
            return reduction_fn(distribution, transform, **kwargs)
    raise NotImplementedError


def _dispatch_reduction(
    distribution_cls: "Type[ds.Distribution]",
    transform_cls: "Type[ds.Transform]",
) -> "List[Tuple[Callable, Callable]]":
    matches = [
        ds.kl._Match(d, t)
        for d, t in _REDUCTION_REGISTRY
        if issubclass(distribution_cls, d) and issubclass(transform_cls, t)
    ]
    reductions = []
    for match in sorted(matches):
        reductions += _REDUCTION_REGISTRY[match.types]
    return reductions


def _check_cat_cat(
    cat_distribution: "CatDistribution", cat_transform: "ds.CatTransform"
) -> "bool":
    return (
        cat_distribution.dim == cat_transform.dim
        and len(cat_distribution.base_dists) == len(cat_transform.transforms)
        and all(
            (
                d.event_shape[cat_distribution.dim]
                if -len(d.event_shape) <= cat_distribution.dim < len(d.event_shape)
                else 1
            )
            == l
            for d, l in zip(cat_distribution.base_dists, cat_transform.lengths)
        )
    )


#####################################################################################################
# Distributions whose support is finite
#####################################################################################################


@register_reduction(
    ds.Binomial,
    ds.Transform,
    lambda d, t: t.domain.event_dim == t.codomain.event_dim == 0,
)
@register_reduction(
    ds.Bernoulli,
    ds.Transform,
    lambda d, t: t.domain.event_dim == t.codomain.event_dim == 0,
)
@register_reduction(
    ds.Categorical,
    ds.Transform,
    lambda d, t: t.domain.event_dim == t.codomain.event_dim == 0,
)
def _reduction_finite_support_univariate(
    distribution: "Union[ds.Binomial, ds.Bernoulli, ds.Categorical]",
    transform: "ds.Transform",
    **kwargs: "Any",
) -> "Finite":
    support = distribution.enumerate_support()
    logits = distribution.log_prob(support).movedim(-1, 0)
    atoms = transform(support.movedim(-1, 0))
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    return Finite(logits=logits, atoms=atoms, validate_args=validate_args)


@register_reduction(
    Finite, ds.Transform, lambda d, t: t.domain.event_dim == t.codomain.event_dim == 0
)
def _reduction_finite_univariate(
    distribution: "Finite", transform: "ds.Transform", **kwargs: "Any"
) -> "Finite":
    atoms = transform(distribution.atoms)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    return Finite(logits=distribution.logits, atoms=atoms, validate_args=validate_args)


@register_reduction(Deterministic, ds.Transform)
def _reduction_deterministic_transform(
    distribution: "Deterministic", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[Deterministic, MaskedDistribution]":
    value = transform(distribution.value)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = Deterministic(value, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


#####################################################################################################
# Distributions whose probability density/mass function is closed under affine transforms
#####################################################################################################


@register_reduction(ds.Cauchy, ds.Transform, lambda d, t: is_affine(t))
@register_reduction(ds.Gumbel, ds.Transform, lambda d, t: is_affine(t))
@register_reduction(ds.Laplace, ds.Transform, lambda d, t: is_affine(t))
@register_reduction(ds.Normal, ds.Transform, lambda d, t: is_affine(t))
def _reduction_loc_scale_affine(
    distribution: "Union[ds.Cauchy, ds.Gumbel, ds.Laplace, ds.Normal]",
    transform: "ds.Transform",
    **kwargs: "Any",
) -> "Union[ds.Cauchy, ds.Gumbel, ds.Laplace, ds.Normal, MaskedDistribution]":
    loc = transform(distribution.loc)
    shift = transform(torch.zeros_like(distribution.scale))
    scale = transform(distribution.scale) - shift
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = type(distribution)(loc, scale, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.StudentT, ds.Transform, lambda d, t: is_affine(t))
def _reduction_student_t_affine(
    distribution: "ds.StudentT", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.StudentT, MaskedDistribution]":
    loc = transform(distribution.loc)
    shift = transform(torch.zeros_like(distribution.scale))
    scale = transform(distribution.scale) - shift
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.StudentT(distribution.df, loc, scale, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Uniform, ds.Transform, lambda d, t: is_affine(t))
def _reduction_uniform_affine(
    distribution: "ds.Uniform", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.Uniform, MaskedDistribution]":
    low = transform(distribution.low)
    high = transform(distribution.high)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Uniform(low, high, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


#####################################################################################################
# Distributions whose probability density/mass function is closed under scaling
#####################################################################################################


@register_reduction(ds.Exponential, ds.Transform, lambda d, t: is_scaling(t))
def _reduction_exponential_scaling(
    distribution: "ds.Exponential", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.Exponential, MaskedDistribution]":
    rate = 1 / transform(1 / distribution.rate)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Exponential(rate, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Gamma, ds.Transform, lambda d, t: is_scaling(t))
def _reduction_gamma_scaling(
    distribution: "ds.Gamma", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.Gamma, MaskedDistribution]":
    rate = 1 / transform(1 / distribution.rate)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Gamma(distribution.concentration, rate, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.HalfCauchy, ds.Transform, lambda d, t: is_scaling(t))
@register_reduction(ds.HalfNormal, ds.Transform, lambda d, t: is_scaling(t))
def _reduction_half_cauchy_half_normal_scaling(
    distribution: "Union[ds.HalfCauchy, ds.HalfNormal]",
    transform: "ds.Transform",
    **kwargs: "Any",
) -> "Union[ds.HalfCauchy, ds.HalfNormal, MaskedDistribution]":
    scale = transform(distribution.scale)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = type(distribution)(scale, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Pareto, ds.Transform, lambda d, t: is_scaling(t))
def _reduction_pareto_scaling(
    distribution: "ds.Pareto", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.Pareto, MaskedDistribution]":
    scale = transform(distribution.scale)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Pareto(scale, distribution.alpha, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Weibull, ds.Transform, lambda d, t: is_scaling(t))
def _reduction_weibull_scaling(
    distribution: "ds.Weibull", transform: "ds.Transform", **kwargs: "Any"
) -> "Union[ds.Weibull, MaskedDistribution]":
    scale = transform(distribution.scale)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Weibull(scale, distribution.concentration, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


#####################################################################################################
# Sum transform (see https://en.wikipedia.org/wiki/List_of_convolutions_of_probability_distributions)
#####################################################################################################


@register_reduction(
    ds.Bernoulli,
    SumTransform,
    lambda d, t: is_identical(d.logits, t.dim - t.domain.event_dim),
)
@register_reduction(
    ds.Geometric,
    SumTransform,
    lambda d, t: is_identical(d.logits, t.dim - t.domain.event_dim),
)
def _reduction_bernoulli_geometric_sum(
    distribution: "Union[ds.Bernoulli, ds.Geometric]",
    transform: "SumTransform",
    **kwargs: "Any",
) -> "Union[ds.Binomial, ds.NegativeBinomial, MaskedDistribution]":
    total_count = distribution.logits.shape[transform.dim - transform.domain.event_dim]
    logits = distribution.logits.select(transform.dim - transform.domain.event_dim, 0)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    distribution_cls = (
        ds.Binomial if isinstance(distribution, ds.Bernoulli) else ds.NegativeBinomial
    )
    base_distribution = distribution_cls(
        total_count, logits=logits, validate_args=validate_args
    )
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(
    ds.Binomial,
    SumTransform,
    lambda d, t: is_identical(d.logits, t.dim - t.domain.event_dim),
)
@register_reduction(
    ds.NegativeBinomial,
    SumTransform,
    lambda d, t: is_identical(d.logits, t.dim - t.domain.event_dim),
)
def _reduction_binomial_negative_binomial_sum(
    distribution: "Union[ds.Binomial, ds.NegativeBinomial]",
    transform: "SumTransform",
    **kwargs: "Any",
) -> "Union[ds.Binomial, ds.NegativeBinomial, MaskedDistribution]":
    total_count = transform(distribution.total_count)
    logits = distribution.logits.select(transform.dim - transform.domain.event_dim, 0)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = type(distribution)(
        total_count, logits=logits, validate_args=validate_args
    )
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Poisson, SumTransform)
def _reduction_poisson_sum(
    distribution: "ds.Poisson", transform: "SumTransform", **kwargs: "Any"
) -> "Union[ds.Poisson, MaskedDistribution]":
    rate = transform(distribution.rate)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Poisson(rate, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Cauchy, SumTransform)
@register_reduction(ds.Normal, SumTransform)
def _reduction_cauchy_normal_sum(
    distribution: "Union[ds.Cauchy, ds.Normal]",
    transform: "SumTransform",
    **kwargs: "Any",
) -> "Union[ds.Cauchy, ds.Normal, MaskedDistribution]":
    loc = transform(distribution.loc)
    scale = transform(distribution.scale ** 2).sqrt()
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = type(distribution)(loc, scale, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(
    ds.Gamma,
    SumTransform,
    lambda d, t: is_identical(d.rate, t.dim - t.domain.event_dim),
)
def _reduction_gamma_sum(
    distribution: "ds.Gamma", transform: "SumTransform", **kwargs: "Any"
) -> "Union[ds.Gamma, MaskedDistribution]":
    concentration = transform(distribution.concentration)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Gamma(concentration, distribution.rate, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


@register_reduction(ds.Chi2, SumTransform)
def _reduction_chi2_sum(
    distribution: "ds.Chi2", transform: "SumTransform", **kwargs: "Any"
) -> "Union[ds.Chi2, MaskedDistribution]":
    df = transform(distribution.df)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.Chi2(df, validate_args)
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


#####################################################################################################
# Special transforms
#####################################################################################################


@register_reduction(ds.Cauchy, ds.AbsTransform, lambda d, t: (d.loc == 0).all())
@register_reduction(ds.Normal, ds.AbsTransform, lambda d, t: (d.loc == 0).all())
def _reduction_cauchy_normal_abs(
    distribution: "Union[ds.Cauchy, ds.Normal]",
    transform: "ds.AbsTransform",
    **kwargs: "Any",
) -> "Union[ds.HalfCauchy, ds.HalfNormal]":
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    distribution_cls = (
        ds.HalfCauchy if isinstance(distribution, ds.Cauchy) else ds.HalfNormal
    )
    return distribution_cls(distribution.scale, validate_args)


@register_reduction(ds.Normal, ds.ExpTransform)
def _reduction_normal_exp(
    distribution: "ds.Normal", transform: "ds.ExpTransform", **kwargs: "Any"
) -> "ds.LogNormal":
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    return ds.LogNormal(distribution.loc, distribution.scale, validate_args)


@register_reduction(ds.Normal, ds.StickBreakingTransform)
def _reduction_normal_stick_breaking(
    distribution: "ds.Normal", transform: "ds.StickBreakingTransform", **kwargs: "Any"
) -> "Union[ds.LogisticNormal, MaskedDistribution]":
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = ds.LogisticNormal(
        distribution.loc, distribution.scale, validate_args
    )
    mask = getattr(transform, "transformed_mask", torch.as_tensor(True))
    return (
        MaskedDistribution(base_distribution, mask, validate_args)
        if not mask.all()
        else base_distribution
    )


#####################################################################################################
# Composite distribution/transform
#####################################################################################################


@register_reduction(ds.Independent, ds.Transform)
def _reduction_independent_transform(
    distribution: "ds.Independent", transform: "ds.Transform", **kwargs: "Any"
) -> "ds.Distribution":
    base_distribution = reduce(distribution.base_dist, transform, **kwargs)
    ndims = len(distribution.event_shape) - transform.domain.event_dim
    ndims += transform.codomain.event_dim - len(base_distribution.event_shape)
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    return (
        ds.Independent(base_distribution, ndims, validate_args)
        if ndims > 0
        else base_distribution
    )


@register_reduction(CatDistribution, ds.CatTransform, _check_cat_cat)
def _reduction_cat_cat(
    distribution: "CatDistribution", transform: "ds.CatTransform", **kwargs: "Any"
) -> "CatDistribution":
    base_distributions = [
        reduce(d, t, **kwargs)
        for d, t in zip(distribution.base_dists, transform.transforms)
    ]
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    return CatDistribution(base_distributions, distribution.dim, validate_args)


@register_reduction(ds.Distribution, ds.CatTransform, lambda d, t: len(t.tseq) == 1)
def _reduction_distribution_cat(
    distribution: "ds.Distribution", transform: "ds.CatTransform", **kwargs: "Any"
) -> "ds.Distribution":
    return reduce(distribution, transform.tseq[0], **kwargs)


@register_reduction(CatDistribution, ds.Transform, lambda d, t: len(d.base_dists) == 1)
def _reduction_cat_transform(
    distribution: "CatDistribution", transform: "ds.Transform", **kwargs: "Any"
) -> "ds.Distribution":
    return reduce(distribution.base_dists[0], transform, **kwargs)


@register_reduction(
    MaskedDistribution,
    ds.Transform,
    lambda d, t: len(d.event_shape) >= t.domain.event_dim,
)
def _reduction_masked_event_transform(
    distribution: "MaskedDistribution", transform: "ds.Transform", **kwargs: "Any"
) -> "MaskedDistribution":
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    base_distribution = reduce(distribution.base_dist, transform, **kwargs)
    return MaskedDistribution(base_distribution, distribution.mask, validate_args)


@register_reduction(
    MaskedDistribution,
    ds.Transform,
    lambda d, t: len(d.event_shape) < t.domain.event_dim and hasattr(t, "mask"),
)
def _reduction_masked_batch_transform(
    distribution: "MaskedDistribution", transform: "ds.Transform", **kwargs: "Any"
) -> "ds.Distribution":
    batch_shape, event_shape = distribution.batch_shape, distribution.event_shape
    mask = distribution.mask[(...,) + (None,) * len(event_shape)].expand(
        batch_shape + event_shape
    )
    transform.mask = (
        transform.mask[(...,) + (None,) * (mask.ndim - transform.mask.ndim)].expand_as(
            mask
        )
        & mask
    )
    return reduce(distribution.base_dist, transform, **kwargs)


# Avoid matching subclasses of `distributions.TransformedDistribution` such as `Gumbel`
@register_reduction(
    ds.TransformedDistribution,
    ds.Transform,
    lambda d, t: type(d) == ds.TransformedDistribution
    or isinstance(d, TransformedDistribution),
)
def _reduction_transformed_transform(
    distribution: "ds.TransformedDistribution",
    transform: "ds.Transform",
    **kwargs: "Any",
) -> "ds.Distribution":
    base_distribution = reduce(
        distribution.base_dist,
        ds.ComposeTransform(distribution.transforms)
        if len(distribution.transforms) > 1
        else distribution.transforms[0],
        **kwargs,
    )
    is_reducible_distribution = not isinstance(
        base_distribution, TransformedDistribution
    )
    return (
        reduce(base_distribution, transform, **kwargs)
        if is_reducible_distribution
        else _reduction_distribution_transform(base_distribution, transform, **kwargs)
    )


@register_reduction(ds.Distribution, ds.IndependentTransform)
def _reduction_distribution_independent(
    distribution: "ds.Distribution",
    transform: "ds.IndependentTransform",
    **kwargs: "Any",
) -> "ds.Distribution":
    return reduce(distribution, transform.base_transform, **kwargs)


@register_reduction(ds.Distribution, ds.ComposeTransform)
def _reduction_distribution_compose(
    distribution: "ds.Distribution", transform: "ds.ComposeTransform", **kwargs: "Any"
) -> "ds.Distribution":
    for i, part in enumerate(transform.parts):
        distribution = reduce(distribution, part, **kwargs)
    return distribution


#####################################################################################################
# Generic distribution/transform
#####################################################################################################


@register_reduction(ds.Distribution, ds.Transform, lambda d, t: not hasattr(d, "mask"))
def _reduction_distribution_transform(
    distribution: "ds.Distribution", transform: "ds.Transform", **kwargs: "Any"
) -> "ds.Distribution":
    validate_args = kwargs.get("validate_args", distribution._validate_args)
    mask = getattr(transform, "mask", torch.as_tensor(True))
    base_distribution = (
        MaskedDistribution(distribution, mask, validate_args)
        if not mask.all()
        else distribution
    )
    is_identity_transform = (
        isinstance(transform, ds.ComposeTransform) and not transform.parts
    )
    return (
        base_distribution
        if is_identity_transform
        else TransformedDistribution(base_distribution, transform, validate_args)
    )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Stochastic agent."""

from typing import Optional, Tuple, Union

import torch
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.agents.agent import Agent
from actorch.registry import register


__all__ = [
    "StochasticAgent",
]


@register
class StochasticAgent(Agent):
    """Agent that returns a stochastic prediction."""

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        num_random_timesteps: "int" = 0,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        num_random_timesteps:
            The number of initial timesteps for which
            a random prediction is returned.
        device:
            The device.

        Raises
        ------
        ValueError
            If `num_random_timesteps` is not in the integer interval [0, inf).

        """
        if num_random_timesteps < 0 or not float(num_random_timesteps).is_integer():
            raise ValueError(
                f"`num_random_timesteps` ({num_random_timesteps}) "
                f"must be in the integer interval [0, inf)"
            )
        self.num_random_timesteps = int(num_random_timesteps)
        self._num_elapsed_timesteps = 0
        self._policy_state = None
        super().__init__(
            policy,
            observation_space,
            action_space,
            is_batched,
            device,
        )

    # override
    def _reset(self, mask: "ndarray") -> "None":
        if self._policy_state is not None:
            self._policy_state[mask] = 0.0

    # override
    def _predict(self, flat_observation: "ndarray") -> "Tuple[ndarray, ndarray]":
        if self._num_elapsed_timesteps < self.num_random_timesteps:
            random_flat_action = self._flat_action_space.sample()
            random_log_prob = self._flat_action_space.log_prob(random_flat_action)
            self._num_elapsed_timesteps += len(flat_observation)
            return random_flat_action, random_log_prob
        return self._stochastic_predict(flat_observation)

    def _stochastic_predict(
        self, flat_observation: "ndarray"
    ) -> "Tuple[ndarray, ndarray]":
        # Add temporal axis
        flat_observation = flat_observation[:, None, ...]
        input = torch.as_tensor(flat_observation, device=self.device)
        with torch.no_grad():
            _, self._policy_state = self.policy(input, self._policy_state)
        prediction = self.policy.distribution.sample()
        log_prob = self.policy.distribution.log_prob(prediction)
        flat_action = self.policy.decode(prediction)
        flat_action, log_prob = (
            flat_action.to("cpu").numpy(),
            log_prob.to("cpu").numpy(),
        )
        # Remove temporal axis
        flat_action, log_prob = flat_action[:, 0, ...], log_prob[:, 0, ...]
        return flat_action, log_prob

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"is_batched: {self.is_batched}, "
            f"num_random_timesteps: {self.num_random_timesteps}, "
            f"device: {self.device})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Finite distribution."""

from typing import Optional

import torch
from torch import Size, Tensor
from torch.distributions import Distribution, OneHotCategorical, constraints
from torch.distributions.constraints import Constraint

from actorch.distributions.constraints import ordered_real_vector, real_set
from actorch.registry import register


__all__ = [
    "Finite",
]


@register
class Finite(Distribution):
    """Distribution defined over an arbitrary finite support.

    References
    ----------
    .. [1] M. G. Bellemare, W. Dabney, and R. Munos. "A Distributional Perspective on
           Reinforcement Learning". In: ICML. 2017, pp. 449-458.
           URL: https://arxiv.org/abs/1707.06887
    .. [2] G. Barth-Maron, M. W. Hoffman, D. Budden, W. Dabney, D. Horgan, D. TB,
           A. Muldal, N. Heess, and T. Lillicrap. "Distributed Distributional
           Deterministic Policy Gradients". In: ICLR. 2018.
           URL: https://arxiv.org/abs/1804.08617

    """

    has_enumerate_support = True
    arg_constraints = {
        "probs": constraints.simplex,
        "logits": constraints.real_vector,
        "atoms": ordered_real_vector,
    }
    is_discrete = True
    """Whether the distribution is discrete."""

    def __init__(
        self,
        probs: "Optional[Tensor]" = None,
        logits: "Optional[Tensor]" = None,
        atoms: "Optional[Tensor]" = None,
        validate_args: "Optional[bool]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        probs:
            The event probabilities.
            Must be ``None`` if `logits` is provided.
        logits:
            The event log probabilities (unnormalized).
            Must be ``None`` if `probs` is provided.
        atoms:
            The atoms that form the support of the distribution, sorted in (strict) ascending order.
            Default to `{0, ..., N - 1}` where `N` is ``probs.shape[-1]`` or ``logits.shape[-1]``.
        validate_args:
            True to validate the arguments, False otherwise.
            Default to ``__debug__``.

        """
        param = probs if probs is not None else logits
        if atoms is None:
            atoms = torch.arange(param.shape[-1], device=param.device)
        atoms = atoms.type(param.type())
        param, self.atoms = torch.broadcast_tensors(param, atoms)
        if probs is not None:
            probs = probs.expand_as(param)
        if logits is not None:
            logits = logits.expand_as(param)
        self._one_hot_categorical = OneHotCategorical(probs, logits, validate_args)
        super().__init__(
            self._one_hot_categorical.batch_shape, validate_args=validate_args
        )

    # override
    def expand(
        self, batch_shape: "Size" = torch.Size(), _instance: "Optional[Finite]" = None
    ) -> "Finite":
        new = self._get_checked_instance(Finite, _instance)
        param_shape = batch_shape + self.atoms.shape[-1:]
        new.atoms = self.atoms.expand(param_shape)
        new._one_hot_categorical = self._one_hot_categorical.expand(batch_shape)
        super(Finite, new).__init__(batch_shape, self.event_shape, False)
        new._validate_args = self._validate_args
        return new

    # override
    @property
    def support(self) -> "Constraint":
        return real_set(self.atoms)

    # override
    @property
    def probs(self) -> "Tensor":
        return self._one_hot_categorical.probs

    # override
    @property
    def logits(self) -> "Tensor":
        return self._one_hot_categorical.logits

    # override
    @property
    def mean(self) -> "Tensor":
        return (self.probs * self.atoms).sum(dim=-1)

    # override
    @property
    def variance(self) -> "Tensor":
        return (self.probs * (self.atoms ** 2)).sum(dim=-1) - self.mean ** 2

    # override
    def sample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        one_hot_sample = self._one_hot_categorical.sample(sample_shape)
        return (self.atoms * one_hot_sample).sum(dim=-1)

    # override
    def log_prob(self, value: "Tensor") -> "Tensor":
        if self._validate_args:
            self._validate_sample(value)
        # Add event dimension
        value = value[..., None].expand(*value.shape, self.atoms.shape[-1])
        expanded_atoms = self.atoms.expand_as(value)
        mask = expanded_atoms == value
        result = (self.logits * mask).sum(dim=-1)
        result[mask.sum(dim=-1) == 0] = -float("inf")
        return result

    # override
    def cdf(self, value: "Tensor") -> "Tensor":
        if self._validate_args:
            self._validate_sample(value)
        # Add event dimension
        value = value[..., None].expand(*value.shape, self.atoms.shape[-1])
        expanded_atoms = self.atoms.expand_as(value)
        mask = expanded_atoms <= value
        return (self.probs * mask).sum(dim=-1)

    # override
    def enumerate_support(self, expand: "bool" = True) -> "Tensor":
        if not (
            self.atoms == self.atoms[(slice(1),) * len(self.batch_shape) + (...,)]
        ).all():
            raise NotImplementedError(
                "`enumerate_support` does not support inhomogeneous atoms"
            )
        values = self.atoms.movedim(-1, 0)
        if not expand:
            values = values[
                (...,)
                + (
                    slice(
                        1,
                    ),
                )
                * len(self.batch_shape)
            ]
        return values

    # override
    def entropy(self) -> "Tensor":
        return self._one_hot_categorical.entropy()

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(probs: {self.probs if self.probs.numel() == 1 else self.probs.shape}, "
            f"atoms: {self.atoms if self.atoms.numel() == 1 else self.atoms.shape})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Maskable transform."""

from abc import ABC, abstractmethod

import torch
from torch import Tensor
from torch.distributions import Transform


__all__ = [
    "MaskableTransform",
]


class MaskableTransform(ABC, Transform):
    """Transform compatible with `actorch.distributions.MaskedDistribution`. When applied
    to a masked distribution, attribute `mask`, i.e. the boolean tensor indicating which
    elements of the transform input are valid (True) and which are not (False), must be
    set accordingly.

    See Also
    --------
    actorch.distributions.masked_distribution.MaskedDistribution

    """

    def __init__(self, cache_size: "int" = 0) -> "None":
        super().__init__(cache_size)
        self.mask = torch.as_tensor(True)[(None,) * self.domain.event_dim]

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(mask: {self.mask if self.mask.numel() == 1 else self.mask.shape})"
        )

    @property
    @abstractmethod
    def transformed_mask(self) -> "Tensor":
        """Return the transformed mask, i.e. the boolean tensor indicating which
        elements of the transform output are valid (True) and which are not (False).

        Returns
        -------
            The transformed mask.

        """
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""(Possibly batched) environment wrappers."""

from actorch.envs.wrappers.flatten_observation import *
from actorch.envs.wrappers.unflatten_action import *
from actorch.envs.wrappers.wrapper import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Transforms."""

from actorch.distributions.transforms.maskable_transform import *
from actorch.distributions.transforms.normalizing_flow import *
from actorch.distributions.transforms.sum_transform import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""TensorBoard loader."""

from typing import Any, Dict, Optional, Sequence, Tuple

import numpy as np
from numpy import ndarray
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator

from actorch.visualizer.loaders.loader import Loader


__all__ = [
    "TensorBoardLoader",
]


class TensorBoardLoader(Loader):
    """Load progress data from TensorBoard progress files."""

    # override
    def load(
        self,
        input_dirpath: "str",
        search_pattern: "str" = ".*",
        exclude_names: "Optional[Sequence[str]]" = None,
        **kwargs: "Any",
    ) -> "Dict[str, Dict[str, Tuple[ndarray, ndarray]]]":
        search_pattern = f"(?=.*events\.out\.tfevents.*)(?={search_pattern})"
        return super().load(input_dirpath, search_pattern, exclude_names, **kwargs)

    # override
    def _load_data(self, filepath: "str", **kwargs: "Any") -> "Dict[str, ndarray]":
        data = {}
        event_accumulator = EventAccumulator(filepath).Reload()
        reference_x = np.empty(0)  # Series with maximum length
        for tag in event_accumulator.Tags()["scalars"]:
            x, y = [], []
            for scalar in event_accumulator.Scalars(tag):
                x.append(float(scalar.step))
                y.append(float(scalar.value))
            x, y = np.array(x), np.array(y)
            if len(x) > len(reference_x):
                reference_x = x
            data[tag] = x, y
        for tag, (x, y) in data.items():
            # Some data series may be shorter because the corresponding values were logged
            # less frequently, therefore they need to be extended to the maximum length
            y_with_gaps = np.full_like(reference_x, np.nan)
            y_with_gaps[np.digitize(x, reference_x) - 1] = y
            data[tag] = y_with_gaps
        return data



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Hydra model."""

from abc import abstractmethod
from typing import Dict, Tuple

from torch import Tensor

from actorch.models.model import Model
from actorch.utils import broadcast_cat


__all__ = [
    "Hydra",
]


class Hydra(Model):
    """Hydra model, consisting of multiple input tails,
    a shared torso and multiple output heads.

      Input 1  ....  Input n
         |              |
     ____|___        ___|____
    | Tail 1 | .... | Tail n |
    |________|      |________|
         |______________|
                 |
              ___|___
             | Torso |
             |_______|
                 |
          _______|______
     ____|___        ___|____
    | Head 1 | .... | Head m |
    |________|      |________|
         |              |
         |              |
      Output 1 .... Output m

    """

    # override
    def _setup(self) -> "None":
        self._setup_tails()
        example_inputs = self.get_example_inputs()
        torso_example_input, _ = self._forward_tails(*example_inputs)
        torso_in_shape = torso_example_input.shape[1:]
        self._setup_torso(torso_in_shape)
        head_example_input, _ = self._forward_torso(
            torso_example_input, *example_inputs[1:]
        )
        head_in_shape = head_example_input.shape[1:]
        self._setup_heads(head_in_shape)

    # override
    def _forward(
        self,
        inputs: "Dict[str, Tensor]",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor]]":
        torso_input, states = self._forward_tails(inputs, states, mask)
        torso_output, states = self._forward_torso(torso_input, states, mask)
        outputs, states = self._forward_heads(torso_output, states, mask)
        return outputs, states

    def _setup_tails(self) -> "None":
        """Setup the model tails."""
        pass

    def _setup_torso(self, in_shape: "Tuple[int, ...]") -> "None":
        """Setup the model torso.

        Parameters
        ----------
        in_shape:
            The input event shape.

        """
        pass

    def _setup_heads(self, in_shape: "Tuple[int, ...]") -> "None":
        """Setup the model heads.

        Parameters
        ----------
        in_shape:
            The input event shape.

        """
        pass

    def _forward_tails(
        self,
        inputs: "Dict[str, Tensor]",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Tensor, Dict[str, Tensor]]":
        """Tails forward pass.

        In the following, let `B = [B_1, ..., B_k]` denote the batch shape,
        and `...` an arbitrary event shape.

        Parameters
        ----------
        inputs:
            The inputs, i.e. a dict whose key-value pairs are consistent
            with the given `in_shapes` initialization argument, shape of
            ``inputs[name]``: ``[*B, *in_shapes[name]]``.
        states:
            The states, i.e. a dict with arbitrary key-value pairs.
            Useful, for example, to store hidden states of recurrent models.
        mask:
            The boolean tensor indicating which batch elements are
            valid (True) and which are not (False), shape: ``[*B]``.

        Returns
        -------
            - The output, shape: ``[*B, *...]``;
            - the possibly updated states.

        """
        return broadcast_cat([*inputs.values()], mask.ndim), states

    def _forward_torso(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Tensor, Dict[str, Tensor]]":
        """Torso forward pass.

        In the following, let `B = [B_1, ..., B_k]` denote the batch shape,
        and `...` an arbitrary event shape.

        Parameters
        ----------
        input:
            The input, shape: ``[*B, *...]``.
        states:
            The states, i.e. a dict with arbitrary key-value pairs.
            Useful, for example, to store hidden states of recurrent models.
        mask:
            The boolean tensor indicating which batch elements are
            valid (True) and which are not (False), shape: ``[*B]``.

        Returns
        -------
            - The output, shape: ``[*B, *...]``;
            - the possibly updated states.

        """
        return input, states

    @abstractmethod
    def _forward_heads(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor]]":
        """Heads forward pass.

        In the following, let `B = [B_1, ..., B_k]` denote the batch shape,
        and `...` an arbitrary event shape.

        Parameters
        ----------
        input:
            The input, shape: ``[*B, *...]``.
        states:
            The states, i.e. a dict with arbitrary key-value pairs.
            Useful, for example, to store hidden states of recurrent models.
        mask:
            The boolean tensor indicating which batch elements are
            valid (True) and which are not (False), shape: ``[*B]``.

        Returns
        -------
            - The outputs, i.e. a dict whose key-value pairs are consistent
              with the given `out_shapes` initialization argument, shape of
              ``outputs[name]``: ``[*B, *out_shapes[name]]``;
            - the possibly updated states.

        """
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Lambda return."""

from typing import Optional, Tuple, Union

import torch
from torch import Tensor
from torch.distributions import Distribution

from actorch.algorithms.value_estimation.vtrace import vtrace


__all__ = [
    "lambda_return",
]


def lambda_return(
    state_values: "Union[Tensor, Distribution]",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    next_state_values: "Optional[Union[Tensor, Distribution]]" = None,
    discount: "float" = 0.99,
    trace_decay: "float" = 1.0,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tuple[Union[Tensor, Distribution], Tensor]":
    """Compute the (distributional) lambda returns, a.k.a. TD(lambda), and
    the corresponding advantages, a.k.a. GAE(lambda), of a trajectory.

    In the following, let `B` denote the batch size and `T` the trajectory maximum length.

    Parameters
    ----------
    state_values:
        The (distributional) state values (`v_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    rewards:
        The rewards (`r_t` in the literature), shape: ``[B, T]``.
    dones:
        The end-of-episode flags, shape: ``[B, T]``.
    mask:
        The boolean tensor indicating which elements (or batch elements
        if distributional) are valid (True) and which are not (False),
        shape: ``[B, T]``.
    next_state_values:
        The (distributional) next state values (`v_t+1` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
        Default to `state_values` from the second timestep on, with a bootstrap value set to 0.
    discount:
        The discount factor (`gamma` in the literature).
    trace_decay:
        The trace-decay parameter (`lambda` in the literature).
    standardize_advantage:
        True to standardize the advantages, False otherwise.
    epsilon:
        The term added to the denominators to improve numerical stability.

    Returns
    -------
        - The (distributional) lambda returns,
          shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``;
        - the corresponding advantages, shape ``[B, T]``.

    References
    ----------
    .. [1] R. S. Sutton and A. G. Barto. "Reinforcement Learning: An Introduction".
           MIT Press, 1998.
           URL: http://incompleteideas.net/sutton/book/ebook/node74.html
    .. [2] J. Schulman, P. Moritz, S. Levine, M. Jordan, and P. Abbeel. "High-Dimensional
           Continuous Control Using Generalized Advantage Estimation". In: ICLR. 2016.
           URL: https://arxiv.org/abs/1506.02438

    """
    return vtrace(
        state_values=state_values,
        rewards=rewards,
        dones=dones,
        mask=mask,
        log_is_weights=torch.zeros_like(mask),
        next_state_values=next_state_values,
        discount=discount,
        num_return_steps=mask.shape[1],
        trace_decay=trace_decay,
        max_is_weight_trace=1.0,
        max_is_weight_delta=1.0,
        max_is_weight_advantage=1.0,
        standardize_advantage=standardize_advantage,
        epsilon=epsilon,
    )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Recurrent model."""

from typing import Any, Callable, Dict, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor, device

from actorch.models.fc import FC
from actorch.models.hydra import Hydra
from actorch.registry import register


__all__ = [
    "Recurrent",
]


@register
class Recurrent(FC):
    """Recurrent model."""

    def __init__(
        self,
        in_shapes: "Dict[str, Tuple[int, ...]]",
        out_shapes: "Dict[str, Tuple[int, ...]]",
        torso_lstm_config: "Optional[Dict[str, Any]]" = None,
        head_fc_bias: "bool" = True,
        head_activation_builder: "Callable[..., nn.Module]" = nn.Identity,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shapes:
            The input event shapes, i.e. a dict that maps names
            of the inputs to their corresponding event shapes.
        out_shapes:
            The output event shapes, i.e. a dict that maps names
            of the outputs to their corresponding event shapes.
        torso_lstm_config:
            The torso LSTM configuration.
            Default to ``{
                "hidden_size": 64,
                "num_layers": 2,
                "bias": True,
                "batch_first": False,
                "dropout": 0.,
                "bidirectional": False,
                "proj_size": 0,
            }``.
        head_fc_bias:
            True to learn an additive bias in the head
            fully connected layer, False otherwise
            (the same for all heads).
        head_activation_builder:
            The head activation builder
            (the same for all heads).

        """
        self.torso_lstm_config = torso_lstm_config or {
            "hidden_size": 64,
            "num_layers": 2,
            "bias": True,
            "batch_first": False,
            "dropout": 0.0,
            "bidirectional": False,
            "proj_size": 0,
        }
        self.head_fc_bias = head_fc_bias
        self.head_activation_builder = head_activation_builder
        Hydra.__init__(self, in_shapes, out_shapes)

    # override
    def _setup_torso(self, in_shape: "Tuple[int, ...]") -> "None":
        self.torso = nn.LSTM(torch.Size(in_shape).numel(), **self.torso_lstm_config)

    # override
    def _forward_torso(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Tensor, Dict[str, Tensor]]":
        batch_shape = mask.shape
        batch_first = self.torso_lstm_config["batch_first"]
        input = input.flatten(start_dim=mask.ndim)
        if mask.ndim < 2:
            mask = mask[..., None] if batch_first else mask[None]
        mask = (
            mask.reshape(-1, mask.shape[-1])
            if batch_first
            else mask.reshape(mask.shape[0], -1)
        )
        input = input.reshape(*mask.shape, -1)
        packed_input = nn.utils.rnn.pack_padded_sequence(
            input,
            mask.sum(dim=-1 if batch_first else 0).to("cpu"),
            batch_first,
            enforce_sorted=False,
        )
        # Copy to avoid side effects
        states = {k: x for k, x in states.items()}
        hidden_state = states.get("hidden_state")
        cell_state = states.get("cell_state")
        lstm_state = None
        if hidden_state is not None and cell_state is not None:
            lstm_state = tuple(
                x.reshape(x.shape[0], -1, x.shape[-1])
                for x in [hidden_state, cell_state]
            )
        packed_output, lstm_state = self.torso(
            packed_input,
            lstm_state,
        )
        output, _ = nn.utils.rnn.pad_packed_sequence(packed_output, batch_first)
        output = output.reshape(*batch_shape, -1)
        states["hidden_state"], states["cell_state"] = tuple(
            x.reshape(
                -1, *(batch_shape[:-1] if batch_first else batch_shape[1:]), x.shape[-1]
            )
            for x in lstm_state
        )
        return output, states

    # override
    def get_example_inputs(
        self,
        batch_shape: "Tuple[int, ...]" = (1,),
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor], Tensor]":
        example_inputs = super().get_example_inputs(batch_shape, device)
        hidden_size = self.torso_lstm_config["hidden_size"]
        num_layers = self.torso_lstm_config["num_layers"]
        num_directions = 2 if self.torso_lstm_config["bidirectional"] else 1
        proj_size = self.torso_lstm_config["proj_size"]
        example_inputs[1]["hidden_state"] = torch.rand(
            (
                num_directions * num_layers,
                *batch_shape,
                proj_size if proj_size > 0 else hidden_size,
            ),
            device=device,
        )
        example_inputs[1]["cell_state"] = torch.rand(
            (num_directions * num_layers, *batch_shape, hidden_size),
            device=device,
        )
        return example_inputs



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Kronecker-factored approximate curvature preconditioner."""

# Adapted from:
# https://github.com/gpauloski/kfac_pytorch/blob/ee8c1ec78db66b3063adf27e3290c7329e6b8a67/kfac/utils.py#L1
# https://github.com/gpauloski/kfac_pytorch/blob/ee8c1ec78db66b3063adf27e3290c7329e6b8a67/kfac/kfac_preconditioner.py#L1

import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, NamedTuple, Optional, OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.cuda.amp import GradScaler
from torch.distributions.kl import _Match as Match
from torch.optim import Optimizer

from actorch.registry import register


__all__ = [
    "KFAC",
]


class _KFACModule(ABC, nn.Module):
    """Wrap a module to make it compatible with K-FAC preconditioner."""

    def __init__(self, module: "nn.Module") -> "None":
        """Initialize the object.

        Parameters
        ----------
        module:
            The module.

        """
        super().__init__()
        self.module = module
        self.a = None  # Module input saved in the forward pass
        self.g = None  # Module gradient with respect to the output saved in the backward pass
        self._A, self._G = None, None  # Fisher factors A and G
        self._dA, self._dG = None, None  # Eigenvalues of A and G
        self._QA, self._QG = None, None  # Eigenvectors of A and G

    # override
    def state_dict(
        self,
        destination: "Optional[OrderedDict[str, Tensor]]" = None,
        prefix: "str" = "",
        keep_vars: "bool" = False,
    ) -> "OrderedDict[str, Tensor]":
        state_dict = super().state_dict(destination, prefix, keep_vars)
        state_dict[prefix + "a"], state_dict[prefix + "g"] = self.a, self.g
        state_dict[prefix + "_A"], state_dict[prefix + "_G"] = self._A, self._G
        state_dict[prefix + "_dA"], state_dict[prefix + "_dG"] = self._dA, self._dG
        state_dict[prefix + "_QA"], state_dict[prefix + "_QG"] = self._QA, self._QG
        return state_dict

    # override
    def load_state_dict(
        self,
        state_dict: "OrderedDict[str, Tensor]",
        strict: "bool" = True,
    ) -> "NamedTuple":
        state_dict = state_dict.copy()
        self.a, self.g = state_dict.pop("a"), state_dict.pop("g")
        self._A, self._G = state_dict.pop("_A"), state_dict.pop("_G")
        self._dA, self._dG = state_dict.pop("_dA"), state_dict.pop("_dG")
        self._QA, self._QG = state_dict.pop("_QA"), state_dict.pop("_QG")
        return super().load_state_dict(state_dict, strict)

    def update_AG(self, decay: "float") -> "None":
        """Update Fisher factors A and G.

        Parameters
        ----------
        decay:
            The exponential moving average decay rate.

        """
        A, G = self._compute_A(self.a), self._compute_G(self.g)
        if self._A is None:
            self._A, self._G = (
                A.new_ones(A.shape[0]).diag(),
                G.new_ones(G.shape[0]).diag(),
            )
            self._dA, self._dG = A.new_zeros(A.shape[0]), G.new_zeros(G.shape[0])
            self._QA, self._QG = A.new_zeros(A.shape), G.new_zeros(G.shape)
        self._update_exp_moving_average(self._A, A, decay)
        self._update_exp_moving_average(self._G, G, decay)

    def update_eigen_AG(self, epsilon: "float") -> "None":
        """Update the eigenvalues and eigenvectors of A and G.

        Parameters
        ----------
        epsilon:
            The term added to the denominators to improve numerical stability.

        """
        self._dA, self._QA = torch.linalg.eigh(self._A, UPLO="U")
        self._dG, self._QG = torch.linalg.eigh(self._G, UPLO="U")
        self._dA *= (self._dA > epsilon).float()
        self._dG *= (self._dG > epsilon).float()

    def get_preconditioned_grad(self, damping: "float") -> "Tensor":
        """Return the module preconditioned flat gradient.

        Parameters
        ----------
        damping:
            The Tikhonov damping parameter.

        Returns
        -------
            The module preconditioned flat gradient.

        """
        v1 = self._QG.t() @ self.grad @ self._QA
        v2 = v1 / (self._dG[:, None, ...] * self._dA[None] + damping)
        v = self._QG @ v2 @ self._QA.t()
        return v

    def _update_exp_moving_average(
        self, current: "Tensor", new: "Tensor", weight: "float"
    ) -> "None":
        if weight == 1.0:
            return
        current *= weight / (1 - weight)
        current += new
        current *= 1 - weight

    @property
    @abstractmethod
    def grad(self) -> "Tensor":
        """Return the module flat gradient.

        Returns
        -------
            The module flat gradient.

        """
        raise NotImplementedError

    @grad.setter
    @abstractmethod
    def grad(self, value: "Tensor") -> "None":
        """Set the module flat gradient.

        Parameters
        ----------
        value:
            The module flat gradient.

        """
        raise NotImplementedError

    @abstractmethod
    def _compute_A(self, a: "Tensor") -> "Tensor":
        """Compute Fisher factor A.

        Parameters
        ----------
        a:
            The module input.

        Returns
        -------
            The Fisher factor A.

        """
        raise NotImplementedError

    @abstractmethod
    def _compute_G(self, g: "Tensor") -> "Tensor":
        """Compute Fisher factor G.

        Parameters
        ----------
        g:
            The module gradient with respect to the output.

        Returns
        -------
            The Fisher factor G.

        """
        raise NotImplementedError


class _KFACLinear(_KFACModule):
    """K-FAC linear layer wrapper."""

    # override
    @property
    def grad(self) -> "Tensor":
        weight_grad = self.module.weight.grad.flatten(start_dim=1)
        if self.module.bias is not None:
            bias_grad = self.module.bias.grad[..., None]
            return torch.cat([weight_grad, bias_grad], dim=1)
        return weight_grad

    # override
    @grad.setter
    def grad(self, value: "Tensor") -> "None":
        if self.module.bias is not None:
            weight_grad = value[:, :-1].reshape_as(self.module.weight)
            bias_grad = value[:, -1:].reshape_as(self.module.bias)
            self.module.bias.grad.copy_(bias_grad)
        else:
            weight_grad = value.reshape_as(self.module.weight)
        self.module.weight.grad.copy_(weight_grad)

    # override
    def _compute_A(self, a: "Tensor") -> "Tensor":
        # Shape of a: [batch_size, in_features]
        if self.module.bias is not None:
            a = torch.cat([a, a.new_ones(a.shape[0], 1)], dim=1)
        return a.t() @ (a / a.shape[0])

    # override
    def _compute_G(self, g: "Tensor") -> "Tensor":
        # Shape of g: [batch_size, out_features]
        return g.t() @ (g / g.shape[0])


class _KFACConvNd(_KFACLinear):
    """K-FAC N-D convolutional layer wrapper.

    References
    ----------
    .. [1] R. Grosse and J. Martens. "A Kronecker-factored approximate Fisher matrix
           for convolution layers". In: ICML. 2016, pp. 573-582.
           URL: https://arxiv.org/abs/1602.01407

    """

    # override
    def _compute_A(self, a: "Tensor") -> "Tensor":
        # Shape of a: [batch_size, in_channels, *in_shape]
        a = self._extract_patches(a)
        spatial_size = a.shape[1:-1].numel()
        a = a.reshape(-1, a.shape[-1])
        if self.module.bias is not None:
            a = torch.cat([a, a.new_ones(a.shape[0], 1)], dim=1)
        a /= spatial_size
        return a.t() @ (a / a.shape[0])

    # override
    def _compute_G(self, g: "Tensor") -> "Tensor":
        # Shape of g: [batch_size, out_channels, *out_shape]
        spatial_size = g.shape[2:].numel()
        g = g.movedim(1, -1)
        g = g.contiguous()
        g = g.reshape(-1, g.shape[-1]) * spatial_size
        return g.t() @ (g / g.shape[0])

    def _extract_patches(self, x: "Tensor") -> "Tensor":
        # Shape of x: [batch_size, in_channels, *in_shape]
        kernel_size = self.module.kernel_size
        stride = self.module.stride
        padding = self.module.padding
        if sum(padding) > 0:
            x = F.pad(x, tuple([x for x in reversed(padding) for _ in range(2)]))
        for i, (size, step) in enumerate(zip(kernel_size, stride)):
            x = x.unfold(2 + i, size, step)
        dim = len(kernel_size) + 1
        x = x.movedim(1, dim).contiguous()
        x = x.reshape(*x.shape[0:dim], x.shape[dim:].numel())
        return x


@register
class KFAC(Optimizer):
    """Kronecker-factored approximate curvature preconditioner.

    Precondition the gradients of a model with a layer-wise Fisher
    information matrix approximation.

    References
    ----------
    .. [1] J. Martens and R. Grosse. "Optimizing Neural Networks with
           Kronecker-factored Approximate Curvature". In: ICML. 2015, pp. 2408-2417.
           URL: https://arxiv.org/abs/1503.05671

    Examples
    --------
    >>> import torch
    >>>
    >>>
    >>> model = torch.nn.Linear(64, 32)
    >>> criterion = torch.nn.MSELoss()
    >>> optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)
    >>> preconditioner = KFAC(model)
    >>> data, target = torch.rand(10, 64), torch.rand(10, 32)
    >>> optimizer.zero_grad()
    >>> output = model(data)
    >>> loss = criterion(output, target)
    >>> loss.backward()
    >>> preconditioner.step()
    >>> optimizer.step()

    """

    _SUPPORTED_MODULE_TYPES = {
        nn.Linear: _KFACLinear,
        nn.Conv1d: _KFACConvNd,
        nn.Conv2d: _KFACConvNd,
        nn.Conv3d: _KFACConvNd,
    }

    def __init__(
        self,
        model: "nn.Module",
        lr: "float" = 0.1,
        factor_decay: "float" = 0.95,
        damping: "float" = 0.001,
        kl_clip: "float" = 0.001,
        factor_update_freq: "int" = 10,
        kfac_update_freq: "int" = 100,
        grad_scaler: "Optional[GradScaler]" = None,
        epsilon: "float" = 1e-6,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        model:
            The model to perform K-FAC updates on.
        lr:
            The learning rate.
        factor_decay:
            The exponential moving average decay rate for the Kronecker factors.
        damping:
            The Tikhonov damping parameter.
        kl_clip:
            The Kullback-Leibler divergence clipping parameter for gradient scaling.
        factor_update_freq:
            The number of iterations between computing and updating the exponential moving
            average of the Kronecker factors.
        kfac_update_freq:
            The number of iterations between applying gradient preconditioning.
        grad_scaler:
            The gradient scaler for AMP training.
        epsilon:
            The term added to the denominators to improve numerical stability.

        Raises
        ------
        ValueError
            If an invalid argument value is provided.

        """
        if lr <= 0.0:
            raise ValueError(f"`lr` ({lr}) must be in the interval (0, inf)")
        if factor_decay > 1.0 or factor_decay <= 0.0:
            raise ValueError(
                f"`factor_decay` ({factor_decay}) must be in the interval (0, 1]"
            )
        if damping <= 0.0:
            raise ValueError(f"`damping` ({damping}) must be in the interval (0, inf)")
        if kl_clip <= 0.0:
            raise ValueError(f"`kl_clip` ({kl_clip}) must be in the interval (0, inf)")
        if factor_update_freq < 1 or not float(factor_update_freq).is_integer():
            raise ValueError(
                f"`factor_update_freq` ({factor_update_freq}) must be in the integer interval [1, inf)"
            )
        if kfac_update_freq < 1 or not float(kfac_update_freq).is_integer():
            raise ValueError(
                f"`kfac_update_freq` ({kfac_update_freq}) must be in the integer interval [1, inf)"
            )
        if epsilon <= 0.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval (0, inf)")
        factor_update_freq = int(factor_update_freq)
        kfac_update_freq = int(kfac_update_freq)
        if kfac_update_freq % factor_update_freq != 0:
            warnings.warn(
                f"`kfac_update_freq` ({kfac_update_freq}) should be a "
                f"multiple of `factor_update_freq` ({factor_update_freq})"
            )

        defaults = {
            "lr": lr,
            "factor_decay": factor_decay,
            "damping": damping,
            "kl_clip": kl_clip,
            "factor_update_freq": factor_update_freq,
            "kfac_update_freq": kfac_update_freq,
            "epsilon": epsilon,
        }
        # K-FAC does not register any parameters
        super().__init__([torch.empty(1)], defaults)
        self.state["grad_scaler"] = grad_scaler
        self._register_submodules(model)

    @property
    def lr(self) -> "float":
        return self.param_groups[0]["lr"]

    @property
    def factor_decay(self) -> "float":
        return self.param_groups[0]["factor_decay"]

    @property
    def damping(self) -> "float":
        return self.param_groups[0]["damping"]

    @property
    def kl_clip(self) -> "float":
        return self.param_groups[0]["kl_clip"]

    @property
    def factor_update_freq(self) -> "int":
        return self.param_groups[0]["factor_update_freq"]

    @property
    def kfac_update_freq(self) -> "int":
        return self.param_groups[0]["kfac_update_freq"]

    @property
    def epsilon(self) -> "float":
        return self.param_groups[0]["epsilon"]

    @property
    def grad_scaler(self) -> "GradScaler":
        return self.state.get("grad_scaler")

    # override
    def state_dict(self) -> "Dict[str, Any]":
        state_dict = super().state_dict()
        state_dict["state"]["grad_scaler"] = None
        if self.grad_scaler:
            state_dict["state"]["grad_scaler"] = self.grad_scaler.state_dict()
        state_dict["state"]["_modules"] = [
            module.state_dict() for module in self._modules.values()
        ]
        return state_dict

    # override
    def load_state_dict(self, state_dict: "Dict[str, Any]") -> "None":
        state_dict = {k: v for k, v in state_dict.items()}  # Copy
        grad_scaler_state = state_dict["state"].pop("grad_scaler")
        if self.grad_scaler and grad_scaler_state:
            self.grad_scaler.load_state_dict(grad_scaler_state)
        module_states = state_dict["state"].pop("_modules")
        if len(module_states) != len(self._modules):
            raise ValueError(
                f"The number of loaded module states ({len(module_states)}) must be "
                f"equal to the number of modules in the model ({len(self._modules)})"
            )
        for module, state in zip(self._modules.values(), module_states):
            module.load_state_dict(state)
        # Workaround to prevent self.grad_scaler
        # and self._modules from being deleted
        state_backup = {k: v for k, v in self.state.items()}  # Copy
        super().load_state_dict(state_dict)
        self.state.update(state_backup)

    # override
    @torch.no_grad()
    def step(self) -> "None":
        update_factor = self._steps % self.factor_update_freq == 0
        update_eigen = self._steps % self.kfac_update_freq == 0

        vs = []
        vg_sum = 0.0
        for module in self._modules.values():
            if update_factor:
                module.update_AG(self.factor_decay)

            if update_eigen:
                module.update_eigen_AG(self.epsilon)

            v = module.get_preconditioned_grad(self.damping)
            vs.append(v)
            vg_sum += (v * module.grad * self.lr ** 2).sum()

        nu = (self.kl_clip / vg_sum).sqrt().clip(max=1.0).item()

        for module, v in zip(self._modules.values(), vs):
            v *= nu
            module.grad = v

        self._steps += 1

    @property
    def _modules(self) -> "Dict[nn.Module, _KFACModule]":
        if "_modules" not in self.state:
            self.state["_modules"] = {}
        return self.state["_modules"]

    @property
    def _steps(self) -> "int":
        return self.state.get("_steps", 0)

    @_steps.setter
    def _steps(self, value: "int") -> "None":
        self.state["_steps"] = value

    def _register_submodules(self, module: "nn.Module") -> "None":
        cls = type(module)
        matches = [Match(t) for t in self._SUPPORTED_MODULE_TYPES if issubclass(cls, t)]
        if matches:
            match = min(matches).types[0]
            self._modules[module] = self._SUPPORTED_MODULE_TYPES[match](module)
            module.register_forward_pre_hook(self._save_input)
            module.register_full_backward_hook(self._save_grad_output)
            return
        submodules = list(module.children())
        if (
            not submodules
            and list(module.parameters())  # Is a leaf module
            and all(
                x.requires_grad for x in module.parameters()
            )  # Has trainable parameters
        ):
            raise NotImplementedError(f"Unsupported module type: {cls}")
        for submodule in submodules:
            self._register_submodules(submodule)

    @torch.no_grad()
    def _save_input(self, module: "nn.Module", input: "Tensor") -> "None":
        if not module.training:
            return
        if self._steps % self.factor_update_freq == 0:
            self._modules[module].a = input[0]

    @torch.no_grad()
    def _save_grad_output(
        self,
        module: "nn.Module",
        grad_input: "Tensor",
        grad_output: "Tensor",
    ) -> "None":
        if not module.training:
            return
        if self._steps % self.factor_update_freq == 0:
            g = grad_output[0]
            if self.grad_scaler is not None:
                g /= self.grad_scaler.get_scale()
            self._modules[module].g = g

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(lr: {self.lr}, "
            f"factor_decay: {self.factor_decay}, "
            f"damping: {self.damping}, "
            f"kl_clip: {self.kl_clip}, "
            f"factor_update_freq: {self.factor_update_freq}, "
            f"kfac_update_freq: {self.kfac_update_freq}, "
            f"grad_scaler: {self.grad_scaler}, "
            f"epsilon: {self.epsilon})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Neural network modules."""

from typing import Any, Tuple

import torch.nn as nn
from torch import Tensor
from torch.nn import Parameter


__all__ = [
    "ParameterModule",
]


class ParameterModule(nn.Module):
    """Wrap a `torch.nn.Parameter` so that it can be included in a
    `torch.nn.Sequential`, `torch.nn.ModuleList` or `torch.nn.ModuleDict`.

    """

    def __init__(self, in_shape: "Tuple[int, ...]", parameter: "Parameter") -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shape:
            The input event shape.
        parameter:
            The parameter to wrap.

        """
        super().__init__()
        self.in_shape = in_shape
        self.parameter = parameter

    # override
    def forward(
        self,
        input: "Tensor",
        *args: "Any",
        **kwargs: "Any",
    ) -> "Tensor":
        """Forward pass.

        Parameters
        ----------
        input:
            The input.

        Returns
        -------
            The underlying parameter, expanded
            to match the input batch shape.

        """
        batch_shape = input.shape[: input.ndim - len(self.in_shape)]
        expanded_parameter = self.parameter.expand(
            batch_shape + (-1,) * self.parameter.ndim
        )
        return expanded_parameter



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Constant schedule."""

from typing import Sequence, Union

import numpy as np
from numpy import ndarray

from actorch.schedules.schedule import Schedule


__all__ = [
    "ConstantSchedule",
]


class ConstantSchedule(Schedule):
    """Constant schedule."""

    def __init__(
        self,
        value: "Union[int, float, Sequence[Union[int, float]], ndarray]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        value:
            The (possibly batched) constant value to return.

        """
        self.value = value
        batch_size = (np.asarray(value).shape or [None])[0]
        super().__init__(batch_size)

    def __call__(self) -> "Union[int, float, ndarray]":
        value = np.array(self.value)  # Copy
        return value if self.batch_size else value.item()

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__}(value: {self.value})"



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Flatten observation (possibly batched) environment wrapper."""

from typing import Any, Tuple, Union

from gym import Env, ObservationWrapper
from numpy import ndarray

from actorch.envs.batched_env import BatchedEnv
from actorch.envs.spaces import Flat
from actorch.envs.utils import Nested
from actorch.envs.wrappers.wrapper import Wrapper


__all__ = [
    "FlattenObservation",
]


class FlattenObservation(Wrapper, ObservationWrapper):
    """(Possibly batched) environment wrapper that flattens the observation."""

    def __init__(self, env: "Union[Env, BatchedEnv]") -> "None":
        super().__init__(env)
        is_batched = isinstance(env.unwrapped, BatchedEnv)
        if is_batched:
            self.single_observation_space = Flat(env.single_observation_space)
        self.observation_space = Flat(env.observation_space, is_batched)
        self._is_observation_flat = hasattr(env.unwrapped, "unflatten_observation")
        if self._is_observation_flat:
            # Disable unflatten_observation to improve performance
            env.unwrapped.unflatten_observation = False

    # override
    def reset(self, *args: "Any", **kwargs: "Any") -> "ndarray":
        observation = self.env.reset(*args, **kwargs)
        return self.observation(observation)

    # override
    def step(self, *args: "Any", **kwargs: "Any") -> "Tuple[ndarray, Any, Any, Any]":
        observation, reward, done, info = self.env.step(*args, **kwargs)
        return self.observation(observation), reward, done, info

    # override
    def observation(self, observation: "Nested") -> "ndarray":
        if self._is_observation_flat:
            return observation
        return self.observation_space.flatten(observation)



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Models."""

from actorch.models.conv import *
from actorch.models.fc import *
from actorch.models.hydra import *
from actorch.models.model import *
from actorch.models.modules import *
from actorch.models.recurrent import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Deterministic distribution."""

from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

import torch
from torch import Size, Tensor
from torch.distributions import Distribution, Independent
from torch.nn import Module

from actorch.distributions import (
    MaskedDistribution,
    NormalizingFlow,
    TransformedDistribution,
)
from actorch.models import Model


__all__ = [
    "Approximator",
]


_T = TypeVar("_T")

_Parameterization = Dict[str, Tuple[str, Tuple[int, ...], Callable[[Tensor], Tensor]]]

_Config = Dict[str, Any]


class Approximator(Distribution, Module):
    """Distribution module."""

    def __init__(
        self,
        input_transforms: "Dict[str, Callable[..., Distribution]]",
        distribution_builders: "Dict[str, Callable[..., Distribution]]",
        distribution_parameterizations: "Dict[str, _Parameterization]",
        distribution_output_transforms: "Dict[str, Callable[..., Distribution]]",
        model_builder: "Callable[..., Model]",
        distribution_configs: "Optional[Dict[str, _Config]]" = None,
        model_config: "Optional[_Config]" = None,
        normalizing_flow_builders: "Optional[Dict[str, Callable[..., NormalizingFlow]]]" = None,
        normalizing_flow_configs: "Optional[Dict[str, _Config]]" = None,
    ) -> "None":
        Module.__init__(self)
        self.input_transforms = input_transforms
        self.distribution_builders = distribution_builders
        self.distribution_parameterizations = distribution_parameterizations
        self.distribution_output_transforms = distribution_output_transforms
        self.model_builder = model_builder
        self.distribution_configs = distribution_configs or {}
        self.model_config = model_config or {}
        self.normalizing_flow_builders = normalizing_flow_builders
        self.normalizing_flow_configs = normalizing_flow_configs or {}
        in_shapes = {
            key: transform.out_shape for key, transform in input_transforms.items()
        }
        out_shapes = {}
        for key, parameterization in distribution_parameterizations.items():
            for param_name in parameterization:
                name, out_shape, _ = parameterization[param_name]
                out_shapes[f"{key}/{name}"] = out_shape
        self._model = model_builder(
            in_shapes,
            out_shapes,
            **self.model_config,
        )
        self._normalizing_flows = {}
        if self.normalizing_flow_builders:
            self._normalizing_flows = {k: None for k in self.normalizing_flow_builders}

    def forward(
        self,
        input: "Tensor",
        state: "Optional[Tensor]" = None,
        mask: "Optional[Tensor]" = None,
    ) -> "Tuple[Tensor, Tensor]":
        model_inputs = {}
        start = stop = 0
        for key, transform in self.input_transforms.items():
            stop += transform.in_shape[0]
            model_inputs[key] = transform(input[..., start:stop])
            start = stop
        model_states = state
        model_outputs, model_states = self._model(model_inputs, model_states, mask)
        self._distributions = {}
        for key, builder in self.distribution_builders.items():
            parameterization = self.distribution_parameterizations[key]
            params = {
                param_name: transform(model_outputs[f"{key}/{name}"])
                for param_name, (name, _, transform) in parameterization.items()
            }
            config = self.distribution_configs.get(key, {})
            distribution = builder(**params, **config)
            reinterpreted_batch_ndims = len(distribution.batch_shape) - mask.ndim
            if reinterpreted_batch_ndims > 0:
                distribution = Independent(distribution, reinterpreted_batch_ndims)
            distribution = MaskedDistribution(distribution, mask)
            if self._normalizing_flows:
                if key in self._normalizing_flows:
                    normalizing_flow = self._normalizing_flows[key]
                    if not normalizing_flow:
                        # First call
                        in_shape = distribution.event_shape
                        config = self.normalizing_flow_configs.get(key, {})
                        normalizing_flow = builder(in_shape, **config)
                        self._normalizing_flow[key] = normalizing_flow
                    distribution = TransformedDistribution(
                        distribution, normalizing_flow
                    )
            self._distributions[key] = distribution



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""(Possibly batched) environment wrapper."""

from typing import Any, Tuple, Union

from gym import Env
from gym import Wrapper as GymWrapper
from gym import spaces

from actorch.envs.batched_env import BatchedEnv


__all__ = [
    "Wrapper",
]


class Wrapper(GymWrapper):
    """(Possibly batched) environment wrapper."""

    def __init__(self, env: "Union[Env, BatchedEnv]") -> "None":
        super().__init__(env)
        self._single_observation_space = None
        self._single_action_space = None

    @property
    def single_observation_space(self) -> "spaces.Space":
        if self._single_observation_space is None:
            return self.env.single_observation_space
        return self._single_observation_space

    @single_observation_space.setter
    def single_observation_space(self, space: "spaces.Space") -> "None":
        self._single_observation_space = space

    @property
    def single_action_space(self) -> "spaces.Space":
        if self._single_action_space is None:
            return self.env.single_action_space
        return self._single_action_space

    @single_action_space.setter
    def single_action_space(self, space: "spaces.Space") -> "None":
        self._single_action_space = space

    # override
    def reset(self, *args: "Any", **kwargs: "Any") -> "Any":
        return self.env.reset(*args, **kwargs)

    # override
    def step(self, *args: "Any", **kwargs: "Any") -> "Tuple[Any, Any, Any, Any]":
        return self.env.step(*args, **kwargs)

    # override
    def seed(self, *args: "Any", **kwargs: "Any") -> "None":
        return self.env.seed(*args, **kwargs)

    # override
    def render(self, *args: "Any", **kwargs: "Any") -> "None":
        return self.env.render(*args, **kwargs)

    # override
    def close(self, *args: "Any", **kwargs: "Any") -> "None":
        return self.env.close(*args, **kwargs)

    def __len__(self) -> "int":
        return self.env.__len__()

    # Implicitly forward all other attributes to self.env
    def __getattr__(self, name: "str") -> "Any":
        return getattr(self.env, name)



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Environments."""

from actorch.envs.batched_env import *
from actorch.envs.parallel_batched_env import *
from actorch.envs.serial_batched_env import *
from actorch.envs.spaces import *
from actorch.envs.utils import Nested
from actorch.envs.wrappers import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Importance sampling."""

from typing import Optional, Tuple, Union

import torch
from torch import Tensor
from torch.distributions import Distribution

from actorch.algorithms.value_estimation.generalized_estimator import (
    generalized_estimator,
)


__all__ = [
    "importance_sampling",
]


def importance_sampling(
    state_values: "Union[Tensor, Distribution]",
    action_values: "Union[Tensor, Distribution]",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    log_is_weights: "Tensor",
    next_state_values: "Optional[Union[Tensor, Distribution]]" = None,
    discount: "float" = 0.99,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tuple[Union[Tensor, Distribution], Tensor]":
    """Compute the (distributional) importance sampling targets, a.k.a. IS, and the corresponding
    advantages of a trajectory.

    In the following, let `B` denote the batch size and `T` the trajectory maximum length.

    Parameters
    ----------
    state_values:
        The (distributional) state values (`v_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    action_values:
        The (distributional) action values (`q_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    rewards:
        The rewards (`r_t` in the literature), shape: ``[B, T]``.
    dones:
        The end-of-episode flags, shape: ``[B, T]``.
    mask:
        The boolean tensor indicating which elements (or batch elements
        if distributional) are valid (True) and which are not (False),
        shape: ``[B, T]``.
    log_is_weights:
        The log importance sampling weights, defined as
        ``target_policy.log_prob(action) - behavior_policy.log_prob(action)``,
        where `target_policy` (`pi` in the literature) is the policy being learned about,
        and `behavior_policy` (`mu` in the literature) the one used to generate the data,
        shape: ``[B, T]``.
    next_state_values:
        The (distributional) next state values (`v_t+1` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
        Default to `state_values` from the second timestep on, with a bootstrap value set to 0.
    discount:
        The discount factor (`gamma` in the literature).
    standardize_advantage:
        True to standardize the advantages, False otherwise.
    epsilon:
        The term added to the denominators to improve numerical stability.

    Returns
    -------
        - The (distributional) importance sampling targets,
          shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``;
        - the corresponding advantages, shape: ``[B, T]``.

    References
    ----------
    .. [1] D. Precup, R. S. Sutton, and S. Singh. "Eligibility traces for off-policy
           policy evaluation". In: ICML. 2000.
           URL: https://scholarworks.umass.edu/cs_faculty_pubs/80/
    .. [2] D. Precup, R. S. Sutton, and S. Dasgupta. "Off-policy temporal-difference
           learning with function approximation". In: ICML. 2001, pp. 417-424.
           URL: https://www.cs.mcgill.ca/~dprecup/publications/PSD-01.pdf

    """
    is_weights = log_is_weights.exp()
    trace_weights = is_weights
    delta_weights = torch.ones_like(is_weights)
    advantage_weights = torch.ones_like(is_weights)
    return generalized_estimator(
        state_values=state_values,
        rewards=rewards,
        dones=dones,
        mask=mask,
        trace_weights=trace_weights,
        delta_weights=delta_weights,
        advantage_weights=advantage_weights,
        action_values=action_values,
        next_state_values=next_state_values,
        discount=discount,
        num_return_steps=mask.shape[1],
        trace_decay=1.0,
        standardize_advantage=standardize_advantage,
        epsilon=epsilon,
    )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Constraints."""

import torch
from torch import Tensor
from torch.distributions import constraints


__all__ = [
    "cat",
    "ordered_real_vector",
    "real_set",
]


class _Cat(constraints.cat):
    """Extended version of `torch.distributions.constraints.cat` that
    implements property `is_discrete` and method `check` correctly.

    """

    # override
    @property
    def is_discrete(self) -> "bool":
        return all(c.is_discrete for c in self.cseq)

    # override
    def check(self, value: "Tensor") -> "Tensor":
        if self.dim < -value.ndim or self.dim >= value.ndim:
            raise IndexError(
                f"`dim` ({self.dim}) must be in the integer interval [-{value.ndim}, {value.ndim})"
            )
        chunks = value.split(self.lengths, dim=self.dim)
        checks = [c.check(chunks[i]) for i, c in enumerate(self.cseq)]
        return torch.stack(checks).all(dim=0)

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__[1:]}"
            f"({self.cseq}, "
            f"dim: {self.dim}, "
            f"lengths: {self.lengths})"
        )


class _OrderedRealVector(constraints.Constraint):
    """Constrain to a real-valued vector whose elements are sorted in (strict) ascending order."""

    event_dim = 1

    # override
    def check(self, value: "Tensor") -> "Tensor":
        return (value[..., 1:] > value[..., :-1]).all(dim=-1)


class _RealSet(constraints.Constraint):
    """Constrain to a set (i.e. no duplicates allowed) of real values."""

    is_discrete = True

    def __init__(self, values: "Tensor") -> "None":
        """Initialize the object.

        Parameters
        ----------
        values:
            The set of real values. All dimensions except for the
            last one are interpreted as batch dimensions.

        Raises
        ------
        ValueError
            If `values` contains duplicates.

        """
        sorted_values = values.sort(dim=-1).values
        if (sorted_values[..., 1:] == sorted_values[..., :-1]).any(dim=-1).any():
            raise ValueError(f"`values` ({values}) must not contain duplicates")
        self._values = values

    # override
    def check(self, value: "Tensor") -> "Tensor":
        # Add event dimension
        value = value[..., None].expand(*value.shape, self._values.shape[-1])
        expanded_support = self._values.expand_as(value)
        return (expanded_support == value).any(dim=-1)

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__[1:]}(values: {self._values})"


# Public interface
cat = _Cat
ordered_real_vector = _OrderedRealVector()
real_set = _RealSet



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Experience replay buffers."""

from actorch.buffers.buffer import *
from actorch.buffers.proportional_buffer import *
from actorch.buffers.rank_based_buffer import *
from actorch.buffers.uniform_buffer import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Registry."""

import functools
from typing import Any, Callable, Optional, Union

from ray import cloudpickle as ray_cloudpickle
from ray.experimental import internal_kv
from ray.tune import Trainable, registry


__all__ = [
    "get",
    "register",
]


_OTHER = (
    "actorch"  # Custom category for registrable components other than tune.Trainable
)


def register(
    component: "Optional[Any]" = None,
    name: "Optional[str]" = None,
) -> "Union[Any, Callable]":
    """Decorator to register a component.

    Parameters
    ----------
    component:
        The component.
    name:
        The name to register the component with.
        Default to ``component.__qualname__``.

    Returns
    -------
        The decorated component.

    """

    @functools.wraps(component)
    def register_fn(component: "Any") -> "Any":
        component_name = name or component.__qualname__
        if hasattr(component, "__mro__") and Trainable in component.__mro__:
            registry.register_trainable(component_name, component)
            return component
        global_registry = registry._global_registry
        global_registry._to_flush[
            (_OTHER, component_name)
        ] = ray_cloudpickle.dumps_debug(component)
        if internal_kv._internal_kv_initialized():
            global_registry.flush_values()
        return component

    if component:
        return register_fn(component)

    return register_fn


def get(name: "str") -> "Any":
    """Return the component previously registered with `name`.

    Parameters
    ----------
    name:
        The component name.

    Returns
    -------
        The component previously registered with `name`.

    Raises
    ------
    KeyError
        If `name` does not exist in ACTorch registry.

    """
    global_registry = registry._global_registry
    if global_registry.contains(registry.TRAINABLE_CLASS, name):
        return global_registry.get(registry.TRAINABLE_CLASS, name)
    if global_registry.contains(_OTHER, name):
        return global_registry.get(_OTHER, name)
    raise KeyError(f"{name}")



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Distribution utilities."""

from functools import singledispatch

import torch
from torch import Tensor, distributions
from torch.distributions import Distribution, Transform


__all__ = [
    "is_affine",
    "is_discrete",
    "is_scaling",
    "l2_project",
]


_DISCRETE_DISTRIBUTIONS = (
    distributions.Bernoulli,
    distributions.Binomial,
    distributions.Categorical,
    distributions.Geometric,
    distributions.Multinomial,
    distributions.NegativeBinomial,
    distributions.OneHotCategorical,
    distributions.Poisson,
)


# Adapted from:
# https://github.com/deepmind/trfl/blob/08ccb293edb929d6002786f1c0c177ef291f2956/trfl/distribution_ops.py#L32
def l2_project(z_p: "Tensor", p: "Tensor", z_q: "Tensor") -> "Tensor":
    """Project distribution `(z_p, p)` onto support `z_q` under L2-metric over CDFs.

    Supports `z_p` and `z_q` are specified as tensors of distinct atoms given in ascending order.
    Let `B = [B_1, ..., B_k]` denote the batch shape, `N_p` the number of atoms in `z_p` and `N_q`
    the number of atoms in `z_q`. This projection works for any support `z_q`. In particular, `N_p`
    does not need to be equal to `N_q`.

    Parameters
    ----------
    z_p:
        The support of distribution `p`, shape: ``[*B, N_p]``.
    p:
        The probability distribution `p(z_p)`, shape: ``[*B, N_p]``.
    z_q:
        The support to project onto, shape: ``[*B, N_q]``.

    Returns
    -------
        The projection of `(z_p, p)` onto support `z_q`
        under Cramer distance, shape: ``[*B, N_q]``.

    References
    ----------
    .. [1] M. G. Bellemare, W. Dabney, and R. Munos. "A Distributional Perspective on
           Reinforcement Learning". In: ICML. 2017, pp. 449-458.
           URL: https://arxiv.org/abs/1707.06887
    .. [2] G. Barth-Maron, M. W. Hoffman, D. Budden, W. Dabney, D. Horgan, D. TB,
           A. Muldal, N. Heess, and T. Lillicrap. "Distributed Distributional
           Deterministic Policy Gradients". In: ICLR. 2018.
           URL: https://arxiv.org/abs/1804.08617

    """
    # Reshape
    z_p = z_p[..., None, :]  # [*B, 1, N_p]
    p = p[..., None, :]  # [*B, 1, N_p]
    z_q = z_q[..., None]  # [*B, N_q, 1]

    # Extract vmin and vmax and construct helper tensors from z_q
    vmin, vmax = z_q[..., :1, :], z_q[..., -1:, :]
    d_pos = torch.cat([z_q, vmin], dim=-2)[..., 1:, :]  # [*B, N_q, 1]
    d_neg = torch.cat([vmax, z_q], dim=-2)[..., :-1, :]  # [*B, N_q, 1]

    # Clip z_p to be in the new support range (vmin, vmax)
    z_p = z_p.min(vmax).max(vmin)  # [*B, 1, N_p]

    # Get the distance between atom values in the support
    d_pos = d_pos - z_q  # z_q[i + 1] - z_q[i], [*B, N_q, 1]
    d_neg = z_q - d_neg  # z_q[i] - z_q[i - 1], [*B, N_q, 1]

    # Ensure that we do not divide by zero, in case of atoms of identical value
    d_neg[d_neg > 0] = 1 / d_neg[d_neg > 0]  # [*B, N_q, 1]
    d_pos[d_pos > 0] = 1 / d_pos[d_pos > 0]  # [*B, N_q, 1]

    delta_qp = z_p - z_q  # [*B, N_q, N_p]
    d_sign = (delta_qp >= 0).type(p.type())  # [*B, N_q, N_p]

    # Matrix of entries sgn(a_ij) * |a_ij|, with a_ij = clip(z_p)[j] - z_q[i]
    delta_hat = (d_sign * delta_qp * d_pos) - (
        (1 - d_sign) * delta_qp * d_neg
    )  # [*B, N_q, N_p]
    q = ((1 - delta_hat).clamp(0, 1) * p).sum(dim=-1)  # [*B, N_q]
    return q


@singledispatch
def is_discrete(distribution: "Distribution") -> "bool":
    """Check if a distribution is discrete.

    Parameters
    ----------
    distribution:
        The distribution.

    Returns
    -------
        True if the distribution is discrete, False otherwise.

    Notes
    -----
    Register a custom distribution type as follows:
    >>> from torch.distributions import Distribution
    >>>
    >>> from actorch.distributions.utils import is_discrete
    >>>
    >>>
    >>> class Custom(Distribution):
    >>>     # Implementation
    >>>
    >>>
    >>> @is_discrete.register(Custom)
    >>> def _is_discrete_custom(distribution):
    >>>     # Implementation

    """
    if hasattr(distribution, "is_discrete"):
        return distribution.is_discrete
    if hasattr(distribution, "base_dist"):
        return is_discrete(distribution.base_dist)
    return isinstance(distribution, _DISCRETE_DISTRIBUTIONS)


@singledispatch
def is_affine(transform: "Transform") -> "bool":
    """Check if a transform is affine.

    Parameters
    ----------
    transform:
        The transform.

    Returns
    -------
        True if the transform is affine, False otherwise.

    Notes
    -----
    Register a custom transform type as follows:
    >>> from torch.distributions import Transform
    >>>
    >>> from actorch.distributions.utils import is_affine
    >>>
    >>>
    >>> class Custom(Transform):
    >>>     # Implementation
    >>>
    >>>
    >>> @is_affine.register(Custom)
    >>> def _is_affine_custom(transform):
    >>>     # Implementation

    """
    if hasattr(transform, "is_constant_jacobian"):
        return transform.is_constant_jacobian
    if isinstance(transform, distributions.ComposeTransform):
        return all(is_affine(t) for t in transform.parts)
    if isinstance(transform, distributions.IndependentTransform):
        return is_affine(transform.base_transform)
    return isinstance(transform, distributions.AffineTransform)


def is_scaling(transform: "Transform") -> "bool":
    """Check if a transform is a scaling.

    Parameters
    ----------
    transform:
        The transform.

    Returns
    -------
        True if the transform is a scaling, False otherwise.

    """
    if not is_affine(transform):
        return False
    shift = transform(torch.zeros((1,) * transform.domain.event_dim))
    return (shift == 0).all().item()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Experience sampler."""

import time
from typing import Any, Callable, Dict, Iterator, Optional, Tuple, Union

from gym import Env

from actorch.agents import Agent
from actorch.envs import BatchedEnv


__all__ = [
    "Sampler",
]


class Sampler:
    """Sampler that enables an agent to interact with
    an environment and sample experiences from it.

    """

    def __init__(
        self,
        env: "Env",
        agent: "Agent",
        callback_fns: "Optional[Dict[str, Callable[..., None]]]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        env:
            The environment to sample experiences from.
        agent:
            The agent that interacts with the environment.
        callback_fns:
            The callback functions, i.e. a dict with the following key-value pairs:
            - "on_episode_start_fn": the function called after each episode start.
                                     It receives as arguments the environment,
                                     the agent, and the sampling statistics;
            - "on_episode_step_fn":  the function called after each episode step.
                                     It receives as arguments the environment,
                                     the agent, the sampling statistics, and the
                                     auxiliary diagnostic information;
            - "on_episode_end_fn":   the function called after each episode end.
                                     It receives as arguments the environment,
                                     the agent, and the sampling statistics.
            Default to ``{}``.

        Raises
        ------
        TypeError
            If `env` is instance of `BatchedEnv`.
        ValueError
            If observation or action spaces of `agent` and `env` are not equal.

        See Also
        --------
        Sampler.stats

        """
        if agent.observation_space != env.observation_space:
            raise ValueError(
                f"Observation spaces of `agent` ({agent.observation_space}) "
                f"and `env` ({env.observation_space}) must be equal"
            )
        if agent.action_space != env.action_space:
            raise ValueError(
                f"Action spaces of `agent` ({agent.action_space}) "
                f"and `env` ({env.action_space}) must be equal"
            )
        self.env = env
        self.agent = agent
        self.callback_fns = callback_fns or {}
        self.reset()
        if isinstance(env, BatchedEnv):
            raise TypeError(f"`env` ({env}) must be instance of {BatchedEnv}")

    @property
    def stats(self) -> "Optional[Dict[str, Any]]":
        """Return the sampling statistics.

        Returns
        -------
            The sampling statistics, i.e. a dict with the following key-value pairs:
            - "num_timesteps":      the number of sampled timesteps;
            - "num_episodes":       the number of sampled episodes;
            - "episode_lengths":    the lengths of each sampled episode;
            - "episode_cumrewards": the cumulative rewards of each sampled episode;
            - "inference_time_ms":  the agent inference times in milliseconds;
            - "step_time_ms":       the environment step times in milliseconds;
            - "render_time_ms":     the environment render times in milliseconds
                                    (if rendering is enabled).

        Notes
        -----
        Callback functions can be used to add custom metrics
        to the sampling statistics.

        """
        return self._stats

    def reset(self) -> "None":
        """Reset the sampler state."""
        self._length = 0
        self._cumreward = 0.0
        self._stats = None
        self._next_observation = None

    def sample(
        self,
        num_timesteps: "Optional[Union[int, float]]" = None,
        num_episodes: "Optional[Union[int, float]]" = None,
        render: "bool" = False,
    ) -> "Iterator[Tuple[Dict[str, Any], bool]]":
        """Sample experiences from the environment for a given
        number of timesteps or episodes.

        Parameters
        ----------
        num_timesteps:
            The number of timesteps to sample.
            Must be ``None`` if `num_episodes` is provided.
        num_episodes:
            The number of episodes to sample.
            Must be ``None`` if `num_timesteps` is provided.
        render:
            True to render the environment, False otherwise.

        Yields
        ------
            The sampled experience, i.e. a dict with the
            following key-value pairs:
            - "observation":      the observation;
            - "action":           the action;
            - "log_prob":         the action log probability;
            - "reward":           the reward;
            - "next_observation": the next observation;
            - "done":             the end-of-episode flag;
            the terminal flag (might differ from `done` if
            the environment has a time limit).

        Raises
        ------
        ValueError
            If both or none of `num_timesteps` and `num_episodes`
            are provided or if `num_timesteps` or `num_episodes`
            are not in the integer interval [1, inf].

        """
        if not (bool(num_timesteps) ^ bool(num_episodes)):
            raise ValueError(
                "Either `num_timesteps` or `num_episodes` must be provided, but not both"
            )
        if num_timesteps is not None and num_timesteps != float("inf"):
            if num_timesteps < 1 or not float(num_timesteps).is_integer():
                raise ValueError(
                    f"`num_timesteps` ({num_timesteps}) must be in the integer interval [1, inf]"
                )
            num_timesteps = int(num_timesteps)
        if num_episodes is not None and num_episodes != float("inf"):
            if num_episodes < 1 or not float(num_episodes).is_integer():
                raise ValueError(
                    f"`num_episodes` ({num_episodes}) must be in the integer interval [1, inf]"
                )
            num_episodes = int(num_episodes)
        num_timesteps = num_timesteps or float("inf")
        num_episodes = num_episodes or float("inf")
        self._stats = {
            "num_timesteps": 0,
            "num_episodes": 0,
            "episode_lengths": [],
            "episode_cumrewards": [],
            "inference_time_ms": [],
            "step_time_ms": [],
        }
        if render:
            self.stats["render_time_ms"] = []
        yield from self._sample(num_timesteps, num_episodes, render)

    def _sample(
        self,
        num_timesteps: "Union[int, float]",
        num_episodes: "Union[int, float]",
        render: "bool",
    ) -> "Iterator[Tuple[Dict[str, Any], bool]]":
        num_sampled_timesteps, num_sampled_episodes = 0, 0
        # Default callbacks
        on_episode_start_fn = self.callback_fns.get("on_episode_start_fn")
        on_episode_step_fn = self.callback_fns.get("on_episode_step_fn")
        on_episode_end_fn = self.callback_fns.get("on_episode_end_fn")
        while (
            num_sampled_timesteps < num_timesteps
            and num_sampled_episodes < num_episodes
        ):
            if self._next_observation is None:
                # First call
                self._next_observation = self.env.reset()
                if render:
                    render_time = time.time()
                    self.env.render()
                    self._stats["render_time_ms"] += (time.time() - render_time) * 1000
                # Callback
                if on_episode_start_fn:
                    on_episode_start_fn(self.env, self.agent, self._stats)
            observation = self._next_observation
            inference_time = time.time()
            action, log_prob = self.agent(observation)
            self._stats["inference_time_ms"] += (time.time() - inference_time) * 1000
            step_time = time.time()
            next_observation, reward, terminal, info = self.env.step(action)
            self._stats["step_time_ms"] += (time.time() - step_time) * 1000
            if render:
                render_time = time.time()
                self.env.render()
                self._stats["render_time_ms"] += (time.time() - render_time) * 1000
            self._next_observation = next_observation
            done = terminal
            self._length += 1
            self._cumreward += reward
            num_sampled_timesteps += 1
            self._stats["num_timesteps"] = num_sampled_timesteps
            # Callback
            if on_episode_step_fn:
                on_episode_step_fn(self.env, self.agent, self._stats, info)

            if terminal:
                self._next_observation = self.env.reset()
                if render:
                    render_time = time.time()
                    self.env.render()
                    self._stats["render_time_ms"] += (time.time() - render_time) * 1000
                self._stats["episode_lengths"].append(self._length)
                self._stats["episode_cumrewards"].append(self._cumreward)
                self._length = 0
                self._cumreward = 0.0
                num_sampled_episodes += 1
                self._stats["num_episodes"] = num_sampled_episodes
                self.agent.reset()
                done = not info.get("TimeLimit.truncated", not terminal)
                # Callback
                if on_episode_end_fn:
                    on_episode_end_fn(self.env, self.agent, self._stats)

            experience = {
                "observation": observation,
                "action": action,
                "log_prob": log_prob,
                "reward": reward,
                "next_observation": next_observation,
                "done": done,
            }
            yield experience, terminal

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(env: {self.env}, "
            f"agent: {self.agent}, "
            f"callback_fns: {self.callback_fns})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Parallel batched environment."""

import multiprocessing as mp
from copy import deepcopy
from typing import Any, Callable, Optional, Sequence, Tuple

import cloudpickle
import numpy as np
from gym import Env
from numpy import ndarray

from actorch.envs.batched_env import BatchedEnv
from actorch.envs.utils import Nested, flatten, unflatten


__all__ = [
    "ParallelBatchedEnv",
]


class _WorkerSharedMemory(mp.Process):
    """Worker process that runs an environment.

    Observations are sent back to the main process via shared memory.

    """

    def __init__(
        self,
        env_builder: "Callable[[], Env]",
        connection: "mp.connection.Connection",
        shared_memory: "mp.RawArray",
        idx: "int",
        daemon: "Optional[bool]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        env_builder:
            The environment builder.
        connection:
            The connection for communicating with the main process.
        shared_memory:
            The shared memory for sending observations back to the
            main process.
        idx:
            The worker index, used to locate the destination in the
            shared memory to write observations to.
        daemon:
            True to run as a daemonic process, False otherwise.

        """
        super().__init__(daemon=daemon)
        self._pickled_env_builder = cloudpickle.dumps(env_builder)
        self._connection = connection
        self._shared_memory = shared_memory
        self._idx = idx

    # override
    def run(self) -> "None":
        env = cloudpickle.loads(self._pickled_env_builder)()
        del self._pickled_env_builder

        # Locate destination in shared memory to write observations to
        dummy_flat_observation = flatten(
            env.observation_space, env.observation_space.sample(), copy=False
        )
        size = dummy_flat_observation.shape[0]
        dtype = dummy_flat_observation.dtype
        del dummy_flat_observation
        destination = np.frombuffer(
            self._shared_memory, dtype, size, (self._idx * size) * dtype.itemsize
        )
        del self._idx

        try:
            while True:
                command, data = self._connection.recv()
                if command == "reset":
                    observation = env.reset()
                    flat_observation = flatten(
                        env.observation_space, observation, copy=False
                    )
                    np.copyto(destination, flat_observation)
                    # Send (result, has_errored) to the main process
                    self._connection.send((None, False))
                elif command == "step":
                    observation, reward, done, info = env.step(data)
                    flat_observation = flatten(
                        env.observation_space, observation, copy=False
                    )
                    np.copyto(destination, flat_observation)
                    self._connection.send(((reward, done, info), False))
                elif command == "seed":
                    self._connection.send((env.seed(data), False))
                elif command == "render":
                    self._connection.send((env.render(**data), False))
                elif command == "close":
                    self._connection.send((None, False))
                    break
                else:
                    raise NotImplementedError(f"Unknown command: {command}")
        except (Exception, KeyboardInterrupt):
            import traceback

            self._connection.send((traceback.format_exc(), True))
        finally:
            env.close()
            self._connection.close()

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__[1:]}()"


class ParallelBatchedEnv(BatchedEnv):
    """Batched environment based on subprocesses."""

    def __init__(
        self, env_builder: "Callable[[], Env]", num_workers: "int" = 1
    ) -> "None":
        super().__init__(env_builder, num_workers)
        dummy_flat_observation = flatten(
            self.single_observation_space,
            self.single_observation_space.sample(),
            copy=False,
        )
        size = len(dummy_flat_observation)
        dtype = dummy_flat_observation.dtype
        typecode = dtype.char
        if typecode == "?":
            from ctypes import c_bool

            typecode = c_bool
        # No locking required, as each child process writes
        # to a non-overlapping portion of the shared memory
        shared_memory = mp.RawArray(typecode, num_workers * size)
        self._parents, self._processes = [], []
        for i in range(num_workers):
            parent, child = mp.Pipe()
            self._parents.append(parent)
            process = _WorkerSharedMemory(
                self.env_builder,
                child,
                shared_memory,
                i,
                daemon=True,
            )
            self._processes.append(process)
            process.start()
            child.close()  # Close after starting the process
        self._observation_buffer = np.frombuffer(shared_memory, dtype).reshape(
            num_workers, -1
        )
        self._reward = np.zeros(num_workers)
        self._done = np.zeros(num_workers, dtype=bool)
        self._info = np.array([{} for _ in range(num_workers)])
        # True to unflatten the observation, False otherwise
        # Used by FlattenObservation wrapper to improve performance
        self.unflatten_observation = True

    # override
    def _reset(self, idx: "Sequence[int]") -> "Nested[ndarray]":
        for i in idx:
            self._parents[i].send(["reset", None])
        errors = []
        for i in idx:
            result, has_errored = self._parents[i].recv()
            if has_errored:
                errors.append((i, result))
        self._raise(errors)
        observation = (
            unflatten(self.observation_space, self._observation_buffer)
            if self.unflatten_observation
            else np.array(self._observation_buffer)
        )  # Copy
        return observation

    # override
    def _step(
        self, action: "Sequence[Nested]", idx: "Sequence[int]"
    ) -> "Tuple[Nested[ndarray], ndarray, ndarray, ndarray]":
        for i in idx:
            self._parents[i].send(["step", action[i]])
        errors = []
        for i in idx:
            result, has_errored = self._parents[i].recv()
            if has_errored:
                errors.append((i, result))
            else:
                self._reward[i], self._done[i], self._info[i] = result
        self._raise(errors)
        observation = (
            unflatten(self.observation_space, self._observation_buffer)
            if self.unflatten_observation
            else np.array(self._observation_buffer)
        )  # Copy
        return (
            observation,
            np.array(self._reward),  # Copy
            np.array(self._done),  # Copy
            deepcopy(self._info),
        )

    # override
    def _seed(self, seed: "Sequence[Optional[int]]") -> "None":
        for i, parent in enumerate(self._parents):
            parent.send(["seed", seed[i]])
        errors = []
        for i, parent in enumerate(self._parents):
            result, has_errored = parent.recv()
            if has_errored:
                errors.append((i, result))
        self._raise(errors)

    # override
    def _render(self, idx: "Sequence[int]", **kwargs: "Any") -> "None":
        for i in idx:
            self._parents[i].send(["render", kwargs])
        errors = []
        for i in idx:
            result, has_errored = self._parents[i].recv()
            if has_errored:
                errors.append((i, result))
        self._raise(errors)

    # override
    def _close(self) -> "None":
        try:
            for parent in self._parents:
                parent.send(["close", None])
            errors = []
            for i, parent in enumerate(self._parents):
                result, has_errored = parent.recv()
                if has_errored:
                    errors.append((i, result))
            self._raise(errors)
        except (BrokenPipeError, OSError):
            pass
        finally:
            self._cleanup()

    def _raise(self, errors: "Sequence[Tuple[int, str]]") -> "None":
        if errors:
            error_msgs = [f"[Worker {i}] {traceback}" for i, traceback in errors]
            self._cleanup()
            raise RuntimeError(f"\n{''.join(error_msgs)}")

    def _cleanup(self) -> "None":
        for parent in self._parents:
            parent.close()
        for process in self._processes:
            if process.is_alive():
                process.terminate()
            process.join()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Decaying left truncated variance bonus agent."""

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from gym.spaces import Space
from numpy import ndarray
from torch import Tensor, device
from torch.distributions import Distribution

from actorch.agents.deterministic_agent import DeterministicAgent
from actorch.agents.stochastic_agent import StochasticAgent
from actorch.distributions import Finite
from actorch.registry import register
from actorch.schedules import LambdaSchedule, Schedule
from actorch.utils import is_identical


__all__ = [
    "DLTVBonusAgent",
    "left_truncated_variance_quantile",
]


def left_truncated_variance_quantile(distribution: "Finite") -> "Tensor":
    """Compute the left truncated variance
    of a quantile distribution.

    Parameters
    ----------
    distribution:
        The quantile distribution.

    Returns
    -------
        The left truncated variance.

    Raises
    ------
    ValueError
        If `distribution.probs` is not identical
        along the last dimension.

    References
    ----------
    .. [1] B. Mavrin, S. Zhang, H. Yao, L. Kong, K. Wu, and Y. Yu.
           "Distributional Reinforcement Learning for Efficient Exploration".
           In: ICML. 2019, pp. 4424-4434.
           URL: https://arxiv.org/abs/1905.06125

    """
    if not is_identical(distribution.probs, dim=-1):
        raise ValueError(
            f"`distribution.probs` ({distribution.probs}) "
            f"must be identical along the last dimension"
        )
    num_atoms = distribution.atoms.shape[-1]
    median_idx = num_atoms // 2
    median = distribution.atoms[..., median_idx]
    left_truncated_variance = (
        (distribution.atoms[..., median_idx:] - median) ** 2
    ).sum(dim=-1) / (2 * num_atoms)
    return left_truncated_variance


@register
class DLTVBonusAgent(StochasticAgent, DeterministicAgent):
    """Agent that adds a decaying left truncated variance bonus to the prediction.

    References
    ----------
    .. [1] B. Mavrin, S. Zhang, H. Yao, L. Kong, K. Wu, and Y. Yu.
           "Distributional Reinforcement Learning for Efficient Exploration".
           In: ICML. 2019, pp. 4424-4434.
           URL: https://arxiv.org/abs/1905.06125

    """

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        suppression_coeff: "Optional[Schedule]" = None,
        left_truncated_variance_fn: "Callable[[Distribution], Tensor]" = None,
        num_random_timesteps: "int" = 0,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        suppression_coeff:
            The suppression coefficient schedule
            (`c_t` in the literature).
            Default to ``LambdaSchedule(
                lambda t: 50 * np.sqrt(np.log(t + 2) / (t + 2))
            )``.
        left_truncated_variance_fn:
            The function that computes the left truncated variance
            of a distribution. This function receives as an argument
            the distribution and returns its left truncated variance.
            Default to ``left_truncated_variance_quantile``.
        num_random_timesteps:
            The number of initial timesteps for which
            a random prediction is returned.
        device:
            The device.

        """
        self.suppression_coeff = suppression_coeff or LambdaSchedule(
            lambda t: 50 * np.sqrt(np.log(t + 2) / (t + 2))
        )
        self.left_truncated_variance_fn = (
            left_truncated_variance_fn or left_truncated_variance_quantile
        )
        StochasticAgent.__init__(
            self,
            policy,
            observation_space,
            action_space,
            is_batched,
            num_random_timesteps,
            device,
        )

    # override
    @property
    def schedules(self) -> "List[Schedule]":
        return [self.suppression_coeff]

    # override
    def _stochastic_predict(
        self, flat_observation: "ndarray"
    ) -> "Tuple[ndarray, ndarray]":
        suppression_coeff = self.suppression_coeff()
        if suppression_coeff < 0.0:
            raise ValueError(
                f"`suppression_coeff` ({suppression_coeff}) must be in the interval [0, inf)"
            )
        # Add temporal axis
        flat_observation = flat_observation[:, None, ...]
        input = torch.as_tensor(flat_observation, device=self.device)
        with torch.no_grad():
            _, self._policy_state = self.policy(input, self._policy_state)
        prediction = self.policy.distribution.deterministic_prediction
        left_truncated_variance = self.left_truncated_variance_fn(
            self.policy.distribution
        )
        bonus = suppression_coeff * left_truncated_variance
        flat_action = self.policy.decode(prediction + bonus).to("cpu").numpy()
        # Remove temporal axis
        flat_action = flat_action[:, 0, ...]
        log_prob = np.zeros(len(flat_action))
        return flat_action, log_prob

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"is_batched: {self.is_batched}, "
            f"suppression_coeff: {self.suppression_coeff}, "
            f"left_truncated_variance_fn: {self.left_truncated_variance_fn}, "
            f"device: {self.device})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Algorithm utilities."""

from contextlib import contextmanager
from typing import Iterator, Tuple

import torch
from torch.nn import Module


__all__ = [
    "count_params",
    "freeze_params",
    "sync_polyak",
]


def count_params(module: "Module") -> "Tuple[int, int]":
    """Return the number of trainable and non-trainable
    parameters in `module`.

    Parameters
    ----------
    module:
        The module.

    Returns
    -------
        The number of trainable parameters;
        the number of non-trainable parameters.

    """
    num_trainable, num_non_trainable = 0, 0
    for param in module.parameters():
        if param.requires_grad:
            num_trainable += param.numel()
        else:
            num_non_trainable += param.numel()
    return num_trainable, num_non_trainable


@contextmanager
def freeze_params(*modules: "Module") -> "Iterator[None]":
    """Context manager that stops the gradient
    flow in `modules`.

    Parameters
    ----------
    modules:
        The modules.

    """
    params = [
        param
        for module in modules
        for param in module.parameters()
        if param.requires_grad
    ]
    try:
        for param in params:
            param.requires_grad = False
        yield
    finally:
        for param in params:
            param.requires_grad = True


def sync_polyak(
    source_module: "Module",
    target_module: "Module",
    polyak_weight: "float" = 0.001,
) -> "None":
    """Synchronize `source_module` with `target_module`
    via Polyak averaging.

    For each `target_param` in `target_module`,
    for each `source_param` in `source_module`:
    ``target_param = (
        (1 - polyak_weight) * target_param +
        polyak_weight * source_param
    )``.

    Parameters
    ----------
    source_module:
        The source module.
    target_module:
        The target module.
    polyak_weight:
        The Polyak weight.

    Raises
    ------
    ValueError
        If `polyak_weight` is not in the interval [0, 1].

    References
    ----------
    .. [1] B. T. Polyak and A. B. Juditsky.
           "Acceleration of Stochastic Approximation by Averaging".
           In: SIAM Journal on Control and Optimization. 1992, pp. 838-855.
           URL: https://doi.org/10.1137/0330046

    """
    if polyak_weight < 0.0 or polyak_weight > 1.0:
        raise ValueError(
            f"`polyak_weight` ({polyak_weight}) must be in the interval [0, 1]"
        )
    if polyak_weight == 0.0:
        return
    if polyak_weight == 1.0:
        target_module.load_state_dict(source_module.state_dict())
        return
    target_params = target_module.parameters()
    source_params = source_module.parameters()
    with torch.no_grad():
        for target_param, source_param in zip(target_params, source_params):
            # Update in-place
            target_param *= (1 - polyak_weight) / polyak_weight
            target_param += source_param
            target_param *= polyak_weight



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Model."""

from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor, device


__all__ = [
    "Model",
]


class Model(ABC, nn.Module):
    """Generic many-to-many model that maps multiple
    inputs to multiple outputs.

      Input 1  ....  Input n
         |              |
     ____|______________|____
    |         Torso          |
    |________________________|
         |              |
         |              |
      Output 1 .... Output m

    """

    def __init__(
        self,
        in_shapes: "Dict[str, Tuple[int, ...]]",
        out_shapes: "Dict[str, Tuple[int, ...]]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shapes:
            The input event shapes, i.e. a dict that maps names
            of the inputs to their corresponding event shapes.
        out_shapes:
            The output event shapes, i.e. a dict that maps names
            of the outputs to their corresponding event shapes.

        """
        super().__init__()
        self.in_shapes = in_shapes
        self.out_shapes = out_shapes
        self._setup()

    # override
    def forward(
        self,
        inputs: "Dict[str, Tensor]",
        states: "Optional[Dict[str, Tensor]]" = None,
        mask: "Optional[Tensor]" = None,
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor]]":
        """Forward pass.

        In the following, let `B = [B_1, ..., B_k]` denote the batch shape.

        Parameters
        ----------
        inputs:
            The inputs, i.e. a dict whose key-value pairs are consistent
            with the given `in_shapes` initialization argument, shape of
            ``inputs[name]``: ``[*B, *in_shapes[name]]``.
        states:
            The states, i.e. a dict with arbitrary key-value pairs.
            Useful, for example, to store hidden states of recurrent models.
            Default to ``{}``.
        mask:
            The boolean tensor indicating which batch elements are
            valid (True) and which are not (False), shape: ``[*B]``.
            Default to ``torch.ones(B, dtype=torch.bool)``.

        Returns
        -------
            - The outputs, i.e. a dict whose key-value pairs are consistent
              with the given `out_shapes` initialization argument, shape of
              ``outputs[name]``: ``[*B, *out_shapes[name]]``;
            - the possibly updated states.

        """
        first_key = list(inputs.keys())[0]
        first_input = inputs[first_key]
        batch_shape = first_input.shape[
            : first_input.ndim - len(self.in_shapes[first_key])
        ]
        if states is None:
            states = {}
        if mask is None:
            mask = torch.ones(batch_shape, dtype=torch.bool, device=first_input.device)
        outputs, states = self._forward(inputs, states, mask)
        return outputs, states

    def _setup(self) -> "None":
        """Setup the model."""
        pass

    @abstractmethod
    def _forward(
        self,
        inputs: "Dict[str, Tensor]",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor]]":
        """See documentation of method `forward'."""
        raise NotImplementedError

    @abstractmethod
    def get_example_inputs(
        self,
        batch_shape: "Tuple[int, ...]" = (1,),
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor], Tensor]":
        """Return a sequence of example inputs for method `forward`.

        Useful, for example, to trace the model via `torch.jit.trace`.

        Parameters
        ----------
        batch_shape:
            The batch shape.
        device:
            The device.

        Returns
        -------
            The example inputs.

        """
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Agent."""

from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.envs import Flat, Nested
from actorch.schedules import Schedule


__all__ = [
    "Agent",
]


class Agent(ABC):
    """Interface between policy and environment that
    optionally implements an exploration strategy.

    """

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        device:
            The device.

        Raises
        ------
        ValueError
            If batch sizes of observation and action are not equal.

        """
        self.policy = policy  # Check consitency withpolicy shapes
        self.observation_space = observation_space
        self.action_space = action_space
        self.is_batched = is_batched
        self.device = device
        self._flat_observation_space = Flat(observation_space, is_batched)
        self._flat_action_space = Flat(action_space, is_batched)
        self._batch_size = None
        if is_batched:
            if (
                self._flat_observation_space.shape[0]
                != self._flat_action_space.shape[0]
            ):
                raise ValueError(
                    f"Batch sizes of observation ({self._flat_observation_space.shape[0]}) "
                    f"and action ({self._flat_action_space.shape[0]}) must be equal"
                )
            self._batch_size = self._flat_observation_space.shape[0]
        self.reset()

    @property
    def schedules(self) -> "List[Schedule]":
        """Return the agent public schedules.

        Returns
        -------
            The agent public schedules.

        """
        return []

    def reset(
        self,
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
    ) -> "None":
        """Reset the (possibly batched) agent state.

        In the following, let `B` denote the batch size.

        Parameters
        ----------
        mask:
            The boolean array indicating which batch elements are
            to reset (True) and which are not (False), shape: ``[B]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(B, dtype=bool)``.

        Raises
        ------
        ValueError
            If length of `mask` is not equal to the batch size.

        """
        batch_size = self._batch_size or 1
        if mask is None:
            mask = np.ones(batch_size, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (batch_size,))
            if len(mask) != batch_size:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be "
                    f"equal to the batch size ({batch_size})"
                )
        return self._reset(mask)

    def __call__(self, observation: "Nested") -> "Tuple[Nested, ndarray]":
        """Map a (possibly batched) observation to a (possibly batched) action.

        In the following, let `B` denote the batch size, `O` the shape of a single
        observation leaf value and `A` the shape of a single action leaf value.

        Parameters
        ----------
        observation:
            The (possibly batched) observation, shape of a leaf value: ``[B, *O]``.

        Returns
        -------
            - The (possibly batched) action, shape of a leaf value: ``[B, *A]``;
            - the (possibly batched) log probability, shape: ``[B]``.

        Raises
        ------
        ValueError
            If batch size of `observation` is not equal to the batch
            size of the given `observation_space` initialization argument.

        """
        flat_observation = self._flat_observation_space.flatten(observation)
        if self._batch_size and len(flat_observation) != self._batch_size:
            raise ValueError(
                f"Batch size of `observation` ({len(flat_observation)}) must "
                f"be equal to the batch size of the given `observation_space` "
                f"initialization argument ({self._batch_size})"
            )
        self.policy.to(self.device)
        if self.policy.training:
            self.policy.eval()
        if not self._batch_size:
            # Add batch axis
            flat_observation = flat_observation[None]
        flat_action, log_prob = self._predict(flat_observation)
        if not self._batch_size:
            # Remove batch axis
            flat_action, log_prob = flat_action[0], log_prob[0]
        action = self._flat_action_space.unflatten(flat_action)
        return action, log_prob

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"is_batched: {self.is_batched}, "
            f"device: {self.device})"
        )

    def _reset(self, mask: "ndarray") -> "None":
        """See documentation of method `reset`."""
        pass

    @abstractmethod
    def _predict(self, flat_observation: "ndarray") -> "Tuple[ndarray, ndarray]":
        """Map a batched flat observation to a batched flat action and its
        corresponding batched log probability.

        In the following, let `B` denote the batch size, `O_f` the size of
        a single flat observation and `A_f` the size of a single flat action.

        Parameters
        ----------
        flat_observation:
            The batched flat observation, shape: ``[B, O_f]``.

        Returns
        -------
            - The batched flat action, shape: ``[B, A_f]``;
            - the batched log probability, shape: ``[B]``.

        """
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Exponential schedule."""

from typing import Any, Dict, Sequence, Union

import numpy as np
from numpy import ndarray

from actorch.schedules.lambda_schedule import LambdaSchedule


__all__ = [
    "ExponentialSchedule",
]


class ExponentialSchedule(LambdaSchedule):
    """Exponential schedule."""

    def __init__(
        self,
        initial_value: "Union[int, float, Sequence[Union[int, float]], ndarray]",
        final_value: "Union[int, float, Sequence[Union[int, float]], ndarray]",
        num_timesteps: "Union[int, Sequence[int], ndarray]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        initial_value:
            The (possibly batched) initial value.
        final_value:
            The (possibly batched) final value.
        num_timesteps:
            The (possibly batched) number of timesteps
            after which the value is set to `final_value`.

        Raises
        ------
        ValueError
            If an invalid argument value is provided.

        """
        try:
            initial_value, final_value, num_timesteps = [
                np.array(v)
                for v in np.broadcast_arrays(
                    initial_value,
                    final_value,
                    num_timesteps,
                )  # Copy
            ]
        except Exception:
            raise ValueError(
                f"`initial_value` ({initial_value}), `final_value` ({final_value}) and "
                f"`num_timesteps` ({num_timesteps}) must be broadcastable to a single shape"
            )
        if ((num_timesteps < 1) | (num_timesteps % 1 != 0)).any():
            raise ValueError(
                f"`num_timesteps` ({num_timesteps}) must be in the integer interval [1, inf)"
            )
        if (initial_value * final_value <= 0.0).any():
            raise ValueError(
                f"`initial_value` ({initial_value}) and `final_value` ({final_value}) "
                f"must be in the interval (-inf, 0) or (0, inf)"
            )
        self.initial_value = initial_value
        self.final_value = final_value
        self.num_timesteps = num_timesteps
        rate = (final_value / initial_value) ** (1 / num_timesteps)
        batch_size = (initial_value.shape or [None])[0]

        def step_fn(
            num_elapsed_timesteps: Union[int, ndarray],
        ) -> "Union[int, float, ndarray]":
            value = np.where(
                num_elapsed_timesteps < num_timesteps,
                initial_value * (rate ** num_elapsed_timesteps),
                final_value,
            )
            return value if batch_size else value.item()

        super().__init__(step_fn, batch_size)

    # override
    def state_dict(self) -> "Dict[str, Any]":
        state_dict = super().state_dict()
        del state_dict["step_fn"]
        return state_dict

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(initial_value: {self.initial_value}, "
            f"final_value: {self.final_value}, "
            f"num_timesteps: {self.num_timesteps})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Distributions."""

from actorch.distributions.cat_distribution import *
from actorch.distributions.constraints import *
from actorch.distributions.deterministic import *
from actorch.distributions.finite import *
from actorch.distributions.masked_distribution import *
from actorch.distributions.registries import *
from actorch.distributions.transformed_distribution import *
from actorch.distributions.transforms import *
